#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Optional, Tuple

import numpy as np
import requests

from bitcoin_safe.gui.qt.custom_edits import QCompleterLineEdit
from bitcoin_safe.gui.qt.dialogs import question_dialog
from bitcoin_safe.gui.qt.util import (
    Message,
    ensure_scheme,
    get_host_and_port,
    read_QIcon,
    remove_scheme,
)
from bitcoin_safe.network_config import (
    NetworkConfig,
    NetworkConfigs,
    get_default_port,
    get_description,
    get_electrum_configs,
    get_esplora_urls,
    get_mempool_url,
)
from bitcoin_safe.pythonbdk_types import BlockchainType, CBFServerType, bdk
from bitcoin_safe.signals import Signals

logger = logging.getLogger(__name__)

import json
import socket
import ssl

from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGroupBox,
    QLabel,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)


def test_mempool_space_server(url: str) -> bool:
    try:
        response = requests.get(f"{url}/api/blocks/tip/height", timeout=2)
        return response.status_code == 200
    except Exception as e:
        logger.warning(f"Mempool.space server connection test failed: {e}")
        return False


def get_electrum_server_version(host: str, port: int, use_ssl: bool = True, timeout=2) -> Optional[str]:
    try:
        # Connect to the server
        with socket.create_connection((host, port), timeout=timeout) as sock:
            # Wrap the socket with SSL if required
            ssock: Optional[socket.socket] = None
            if use_ssl:
                context = ssl.create_default_context()
                ssock = context.wrap_socket(sock, server_hostname=host)
            else:
                ssock = sock

            # Prepare and send the JSON-RPC request
            request = json.dumps({"id": 1, "method": "server.version", "params": ["1.4", "1.4"]}) + "\n"
            ssock.sendall(request.encode())

            # Receive the response
            response = ssock.recv(4096).decode()  # Assuming the response won't exceed 4096 bytes
            response_json = json.loads(response.split("\n")[0])  # Handling potential extra newlines

            # Close the SSL socket if used
            if use_ssl:
                ssock.close()

            # Check and print the server version
            if "result" in response_json:
                logger.debug(f"Server version: {response_json['result']}")
                return response_json["result"]
            else:
                logger.debug(f"Failed to retrieve server version of {host , port , use_ssl}.")
                return None
    except Exception as e:
        logger.debug(f"Connection or communication error: {e}")
        return None


def test_connection(network_config: NetworkConfig) -> Optional[str]:
    if network_config.server_type == BlockchainType.Electrum:
        try:
            host, port = get_host_and_port(network_config.electrum_url)

            if host is None or port is None:
                logger.warning(f"No host or port given")
                return None
            return get_electrum_server_version(host=host, port=port, use_ssl=network_config.electrum_use_ssl)
        except Exception as e:
            logger.warning(f"Electrum connection test failed: {e}")
            return None

    elif network_config.server_type == BlockchainType.Esplora:
        try:
            # Assuming Esplora's REST API for testing connection
            response = requests.get(f"{network_config.esplora_url}/blocks/tip/height", timeout=2)
            if response.status_code == 200:
                return response.json()
            else:
                return None

        except Exception as e:
            logger.warning(f"Esplora API connection test failed: {e}")
            return None

    elif network_config.server_type == BlockchainType.RPC:
        try:
            # Assuming Bitcoin RPC interface for testing connection
            headers = {"content-type": "application/json"}
            payload = {"jsonrpc": "1.0", "id": "curltest", "method": "getblockchaininfo", "params": []}
            response = requests.post(
                f"{network_config.rpc_ip}:{network_config.rpc_port}",
                json=payload,
                headers=headers,
                auth=(network_config.rpc_username, network_config.rpc_password),
                timeout=2,
            )
            if response.status_code == 200 and "result" in response.json():
                return response.json()
            else:
                return None
        except Exception as e:
            logger.warning(f"RPC connection test failed: {e}")
            return None

    elif network_config.server_type == BlockchainType.CompactBlockFilter:
        # This case might require a different approach depending on how you intend to connect to the p2p network.
        # This is a placeholder as testing p2p connections is more complex and out of scope for this example.
        raise Exception("Not implemented yet")
    raise Exception(f"Invalud {network_config.server_type}")


class NetworkSettingsUI(QDialog):
    signal_apply_and_restart = pyqtSignal(bdk.Network)
    signal_apply_and_shutdown = pyqtSignal(bdk.Network)
    signal_cancel = pyqtSignal()

    def __init__(
        self, network: bdk.Network, network_configs: NetworkConfigs, signals: Optional[Signals], parent=None
    ):
        super().__init__(parent)
        self.signals = signals
        self.network_configs = network_configs
        self._layout = QVBoxLayout(self)

        self.setWindowIcon(read_QIcon("logo.svg"))
        self.network_combobox = QComboBox(self)
        for _network in bdk.Network:
            self.network_combobox.addItem(
                read_QIcon(f"bitcoin-{_network.name.lower()}.svg"), _network.name, userData=_network
            )
        self._layout.addWidget(self.network_combobox)

        self.groupbox_connection = QGroupBox(parent=self)
        self._layout.addWidget(self.groupbox_connection)
        self.groupbox_connection_layout = QVBoxLayout(self.groupbox_connection)

        self.server_type_comboBox = QComboBox(self)
        for blockchain_type in BlockchainType.active_types():
            self.server_type_comboBox.addItem(BlockchainType.to_text(blockchain_type))

        self.groupbox_connection_layout.addWidget(self.server_type_comboBox)

        self.stackedWidget = QStackedWidget(self)
        self.groupbox_connection_layout.addWidget(self.stackedWidget)

        # Compact Block Filters
        self.compactBlockFiltersTab = QWidget()
        self.compactBlockFiltersLayout = QFormLayout(self.compactBlockFiltersTab)
        self.compactBlockFiltersLayout.setFieldGrowthPolicy(
            QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow
        )

        self.cbf_server_typeComboBox = QComboBox(self.compactBlockFiltersTab)
        self.cbf_server_typeComboBox.addItem(self.tr("Manual"), CBFServerType.Manual)
        self.cbf_server_typeComboBox.addItem(self.tr("Automatic"), CBFServerType.Automatic)
        self.cbf_server_typeComboBox.setCurrentIndex(1)

        self.cbf_server_typeComboBox_label = QLabel()
        self.compactBlockFiltersLayout.addRow(
            self.cbf_server_typeComboBox_label, self.cbf_server_typeComboBox
        )

        self.compactblockfilters_ip_address_edit = QCompleterLineEdit(network=network)
        self.compactblockfilters_ip_address_edit.setEnabled(False)
        self.compactblockfilters_ip_address_edit_label = QLabel()
        self.compactBlockFiltersLayout.addRow(
            self.compactblockfilters_ip_address_edit_label, self.compactblockfilters_ip_address_edit
        )

        self.compactblockfilters_port_edit = QCompleterLineEdit(
            network=network,
            suggestions={
                network: [str(get_default_port(network, server_type=BlockchainType.CompactBlockFilter))]
                for network in bdk.Network
            },
        )
        self.compactblockfilters_port_edit.setEnabled(False)
        self.compactblockfilters_port_edit_label = QLabel()
        self.compactBlockFiltersLayout.addRow(
            self.compactblockfilters_port_edit_label, self.compactblockfilters_port_edit
        )

        self.cbf_description = QLabel()
        self.cbf_description.setWordWrap(True)
        self.cbf_description.setTextFormat(Qt.TextFormat.RichText)
        self.cbf_description.setOpenExternalLinks(True)  # Enable opening links
        self.compactBlockFiltersLayout.addRow("", self.cbf_description)

        self.stackedWidget.addWidget(self.compactBlockFiltersTab)

        # Electrum Server
        self.electrumServerTab = QWidget()
        self.electrumServerLayout = QFormLayout(self.electrumServerTab)
        self.electrumServerLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        #
        self.electrum_url_edit = QCompleterLineEdit(
            network=network,
            suggestions={
                network: list(
                    np.unique(
                        [electrum_config.url for electrum_config in get_electrum_configs(network).values()]
                    )
                )
                for network in bdk.Network
            },
        )
        self.electrum_url_edit.editingFinished.connect(self.on_electrum_url_editing_finished)
        self.electrum_use_ssl_checkbox = QCheckBox()

        self.electrum_url_edit_url_label = QLabel()
        self.electrumServerLayout.addRow(self.electrum_url_edit_url_label, self.electrum_url_edit)
        self.electrum_use_ssl_checkbox_label = QLabel()
        self.electrumServerLayout.addRow(self.electrum_use_ssl_checkbox_label, self.electrum_use_ssl_checkbox)

        self.electrum_description = QLabel()
        self.electrum_description.setWordWrap(True)
        self.electrum_description.setTextFormat(Qt.TextFormat.RichText)
        self.electrum_description.setOpenExternalLinks(True)  # Enable opening links
        self.electrumServerLayout.addRow("", self.electrum_description)

        self.stackedWidget.addWidget(self.electrumServerTab)

        # Esplora Server
        self.esploraServerTab = QWidget()
        self.esploraServerLayout = QFormLayout(self.esploraServerTab)
        self.esploraServerLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.esplora_url_edit = QCompleterLineEdit(
            network=network,
            suggestions={
                network: list(np.unique(list(get_esplora_urls(network).values()))) for network in bdk.Network
            },
        )

        self.esplora_url_edit_label = QLabel()
        self.esploraServerLayout.addRow(self.esplora_url_edit_label, self.esplora_url_edit)

        self.esplora_description = QLabel()
        self.esplora_description.setWordWrap(True)
        self.esplora_description.setTextFormat(Qt.TextFormat.RichText)
        self.esplora_description.setOpenExternalLinks(True)  # Enable opening links
        self.esploraServerLayout.addRow("", self.esplora_description)

        self.stackedWidget.addWidget(self.esploraServerTab)

        # RPC
        self.rpcTab = QWidget()
        self.rpcTabLayout = QFormLayout(self.rpcTab)
        self.rpcTabLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)

        self.rpc_ip_address_edit = QCompleterLineEdit(network=network)
        self.rpc_port_edit = QCompleterLineEdit(
            network=network,
            suggestions={
                network: [str(get_default_port(network, server_type=BlockchainType.CompactBlockFilter))]
                for network in bdk.Network
            },
        )
        self.rpc_username_edit = QCompleterLineEdit(network=network)
        self.rpc_password_edit = QCompleterLineEdit(network=network)

        self.rpc_ip_address_edit_label = QLabel()
        self.rpcTabLayout.addRow(self.rpc_ip_address_edit_label, self.rpc_ip_address_edit)
        self.rpc_port_edit_label = QLabel()
        self.rpcTabLayout.addRow(self.rpc_port_edit_label, self.rpc_port_edit)
        self.rpc_username_edit_label = QLabel()
        self.rpcTabLayout.addRow(self.rpc_username_edit_label, self.rpc_username_edit)
        self.rpc_password_edit_label = QLabel()
        self.rpcTabLayout.addRow(self.rpc_password_edit_label, self.rpc_password_edit)

        self.rpc_description = QLabel()
        self.rpc_description.setWordWrap(True)
        self.rpc_description.setTextFormat(Qt.TextFormat.RichText)
        self.rpc_description.setOpenExternalLinks(True)  # Enable opening links
        self.rpcTabLayout.addRow("", self.rpc_description)

        self.stackedWidget.addWidget(self.rpcTab)

        self.groupbox_blockexplorer = QGroupBox()
        self.groupbox_blockexplorer_layout = QVBoxLayout(self.groupbox_blockexplorer)
        self.edit_mempool_url = QCompleterLineEdit(
            network=network,
            suggestions={network: list(get_mempool_url(network).values()) for network in bdk.Network},
        )
        self.groupbox_blockexplorer_layout.addWidget(self.edit_mempool_url)
        self._layout.addWidget(self.groupbox_blockexplorer)

        # Create buttons and layout
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Help
            | QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel
        )
        self.button_box.accepted.connect(self.on_apply_click)
        self.button_box.rejected.connect(self.on_cancel_click)

        if help_button := self.button_box.button(QDialogButtonBox.StandardButton.Help):
            help_button.setText(self.tr("Test Connection"))
        self.button_box.helpRequested.connect(self.test_connection)

        self._layout.addWidget(self.button_box)

        # Signals and Slots
        self.network_combobox.currentIndexChanged.connect(self.on_network_change)
        self.server_type_comboBox.currentIndexChanged.connect(self.set_server_type_comboBox)
        self.cbf_server_typeComboBox.currentIndexChanged.connect(self.enableIPPortLineEdit)

        self.original_network = network
        self.set_ui(network_configs.configs[network.name])

        self._edits_set_network(self.network)
        if self.signals:
            self.signals.language_switch.connect(self.updateUi)
        self.updateUi()

    def updateUi(self):
        self.setWindowTitle(self.tr("Network Settings"))
        self.groupbox_connection.setTitle(self.tr("Blockchain data source"))
        self.electrum_use_ssl_checkbox.setText(self.tr("Enable SSL"))
        self.esplora_url_edit_label.setText(self.tr("URL:"))
        self.electrum_url_edit_url_label.setText(self.tr("URL:"))
        self.electrum_use_ssl_checkbox_label.setText(self.tr("SSL:"))
        self.compactblockfilters_port_edit_label.setText(self.tr("Port:"))
        self.cbf_server_typeComboBox_label.setText(self.tr("Mode:"))
        self.compactblockfilters_ip_address_edit_label.setText(self.tr("IP Address:"))
        self.rpc_ip_address_edit_label.setText(self.tr("IP Address:"))
        self.rpc_port_edit_label.setText(self.tr("Port:"))
        self.rpc_username_edit_label.setText(self.tr("Username:"))
        self.rpc_password_edit_label.setText(self.tr("Password:"))
        self.groupbox_blockexplorer.setTitle(self.tr("Mempool Instance URL"))
        if ok_button := self.button_box.button(QDialogButtonBox.StandardButton.Ok):
            ok_button.setText(self.tr("Apply && Shutdown"))

    def on_electrum_url_editing_finished(self):
        def get_use_ssl(url: str):
            for electrum_config in get_electrum_configs(self.network).values():
                if url.strip() == electrum_config.url.strip():
                    return electrum_config.use_ssl
            return None

        use_ssl = get_use_ssl(self.electrum_url_edit.text())
        if use_ssl is None:
            return
        logger.debug(f"set use_ssl = {use_ssl}")
        self.electrum_use_ssl = use_ssl

    def _test_connection(self, network_config: NetworkConfig) -> Tuple[str | None, bool]:
        server_connection = test_connection(network_config=network_config)

        mempool_server = test_mempool_space_server(url=network_config.mempool_url)
        return server_connection, mempool_server

    def _format_test_responses(
        self, network_config: NetworkConfig, server_connection: str | None, mempool_server: bool
    ) -> str:
        def format_status(response):
            return "Success" if response else "Failed"

        return self.tr("Responses:\n    {name}: {status}\n    Mempool Instance: {server}").format(
            name=network_config.server_type.name,
            status=format_status(server_connection),
            server=format_status(mempool_server),
        )

    def test_connection(self):
        new_network_config = self.get_network_settings_from_ui()
        server_connection, mempool_server = self._test_connection(network_config=new_network_config)

        Message(self._format_test_responses(new_network_config, server_connection, mempool_server))

    def set_server_type_comboBox(self, new_index: int):
        if self.server_type_comboBox.itemText(new_index) == BlockchainType.to_text(
            BlockchainType.CompactBlockFilter
        ):
            self.stackedWidget.setCurrentWidget(self.compactBlockFiltersTab)
        elif self.server_type_comboBox.itemText(new_index) == BlockchainType.to_text(BlockchainType.Electrum):
            self.stackedWidget.setCurrentWidget(self.electrumServerTab)
        elif self.server_type_comboBox.itemText(new_index) == BlockchainType.to_text(BlockchainType.Esplora):
            self.stackedWidget.setCurrentWidget(self.esploraServerTab)
        elif self.server_type_comboBox.itemText(new_index) == BlockchainType.to_text(BlockchainType.RPC):
            self.stackedWidget.setCurrentWidget(self.rpcTab)

    def on_network_change(self, new_index: int):
        new_network = self.network_combobox.itemData(new_index)

        self._edits_set_network(new_network)
        self.set_ui(self.network_configs.configs[new_network.name])

    def _edits_set_network(self, network: bdk.Network):
        self.compactblockfilters_ip_address_edit.set_network(network)
        self.electrum_url_edit.set_network(network)
        self.esplora_url_edit.set_network(network)
        self.rpc_ip_address_edit.set_network(network)
        self.rpc_username_edit.set_network(network)
        self.rpc_password_edit.set_network(network)
        self.edit_mempool_url.set_network(network)

    def add_to_completer_memory(self):
        self.compactblockfilters_ip_address_edit.add_current_to_memory()
        self.electrum_url_edit.add_current_to_memory()
        self.esplora_url_edit.add_current_to_memory()
        self.rpc_ip_address_edit.add_current_to_memory()
        self.rpc_username_edit.add_current_to_memory()
        self.rpc_password_edit.add_current_to_memory()
        self.edit_mempool_url.add_current_to_memory()

    def on_apply_click(self):
        new_network_config = self.get_network_settings_from_ui()
        server_connection, mempool_server = self._test_connection(network_config=new_network_config)

        if not all([server_connection, mempool_server]):
            if not question_dialog(
                self.tr("Error in server connection.\n{responses}\n\n Do you want to proceed anyway?").format(
                    responses=self._format_test_responses(
                        new_network_config, server_connection, mempool_server
                    )
                )
            ):
                return

        new_network = self.network

        self.add_to_completer_memory()

        self.network_configs.configs[self.network.name] = new_network_config

        self.close()
        self.signal_apply_and_shutdown.emit(new_network)

    def on_cancel_click(self):
        self.set_ui(self.network_configs.configs[self.original_network.name])
        self.signal_cancel.emit()
        self.close()

    # Override keyPressEvent method
    def keyPressEvent(self, event: QKeyEvent | None):
        # Check if the pressed key is 'Esc'
        if event and event.key() == Qt.Key.Key_Escape:
            # Close the widget
            self.on_cancel_click()

        super().keyPressEvent(event)

    def get_network_settings_from_ui(self) -> NetworkConfig:
        "returns current ui as NetworkConfig"
        network_config = NetworkConfig(network=self.network)
        for name in vars(network_config):
            if name.startswith("_"):
                continue
            if not hasattr(self, name):
                logger.error(f"get_network_settings_from_ui: {name} not present in {self.__class__.__name__}")
                continue
            setattr(network_config, name, getattr(self, name))

        return network_config

    def set_ui(self, network_config: NetworkConfig):
        "Sets the ui from a NetworkConfig"

        for name in vars(network_config):
            if name.startswith("_"):
                continue
            if not hasattr(self, name):
                logger.error(f"set_ui: {name} not present in {self.__class__.__name__}")
                continue
            setattr(self, name, getattr(network_config, name))

        self.set_server_type_comboBox(self.server_type_comboBox.currentIndex())
        self._edits_set_network(self.network)

        # set the descriptions
        self.cbf_description.setText(
            get_description(network=self.network, server_type=BlockchainType.CompactBlockFilter)
        )
        self.rpc_description.setText(get_description(network=self.network, server_type=BlockchainType.RPC))
        self.esplora_description.setText(
            get_description(network=self.network, server_type=BlockchainType.Esplora)
        )
        self.electrum_description.setText(
            get_description(network=self.network, server_type=BlockchainType.Electrum)
        )

    def enableIPPortLineEdit(self, index: int):
        if self.cbf_server_typeComboBox.itemData(index) == CBFServerType.Manual:
            self.compactblockfilters_ip_address_edit.setEnabled(True)
            self.compactblockfilters_port_edit.setEnabled(True)
        else:
            self.compactblockfilters_ip_address_edit.setEnabled(False)
            self.compactblockfilters_port_edit.setEnabled(False)

    # Properties for all user entries

    @property
    def mempool_url(self) -> str:
        url = self.edit_mempool_url.text()
        url = url if url.endswith("/") else f"{url}/"
        url = url.replace("api/", "") if url.endswith("api/") else url
        return ensure_scheme(url)

    @mempool_url.setter
    def mempool_url(self, value: str):
        self.edit_mempool_url.setText(value)

    @property
    def network(self) -> bdk.Network:
        return self.network_combobox.currentData()

    @network.setter
    def network(self, value: bdk.Network):
        self.network_combobox.setCurrentText(value.name)

    @property
    def server_type(self) -> BlockchainType:
        return BlockchainType.from_text(self.server_type_comboBox.currentText())

    @server_type.setter
    def server_type(self, server_type: BlockchainType):
        self.server_type_comboBox.setCurrentText(BlockchainType.to_text(server_type))

    @property
    def cbf_server_type(self) -> CBFServerType:
        return self.cbf_server_typeComboBox.currentData()

    @cbf_server_type.setter
    def cbf_server_type(self, cbf_server_type: CBFServerType):
        index = self.cbf_server_typeComboBox.findData(cbf_server_type)
        if index != -1:
            self.cbf_server_typeComboBox.setCurrentIndex(index)
            self.enableIPPortLineEdit(index)

    @property
    def compactblockfilters_ip(self) -> str:
        return self.compactblockfilters_ip_address_edit.text()

    @compactblockfilters_ip.setter
    def compactblockfilters_ip(self, ip):
        self.compactblockfilters_ip_address_edit.setText(ip if ip else "")

    @property
    def compactblockfilters_port(self) -> int:
        try:
            return int(self.compactblockfilters_port_edit.text())
        except:
            return self.network_configs.configs[self.network.name].compactblockfilters_port

    @compactblockfilters_port.setter
    def compactblockfilters_port(self, port):
        self.compactblockfilters_port_edit.setText(str(port))

    @property
    def electrum_url(self) -> str:
        text = self.electrum_url_edit.text()
        return remove_scheme(text)

    @electrum_url.setter
    def electrum_url(self, url: str):
        self.electrum_url_edit.setText(url if url else "")

    @property
    def electrum_use_ssl(self) -> bool:
        return self.electrum_use_ssl_checkbox.isChecked()

    @electrum_use_ssl.setter
    def electrum_use_ssl(self, value: bool):
        self.electrum_use_ssl_checkbox.setChecked(value)

    @property
    def esplora_url(self) -> str:
        url = self.esplora_url_edit.text()
        if "//" not in url:
            url = "http://" + url
        return url

    @esplora_url.setter
    def esplora_url(self, url: str):
        self.esplora_url_edit.setText(url if url else "")

    @property
    def rpc_ip(self) -> str:
        return self.rpc_ip_address_edit.text()

    @rpc_ip.setter
    def rpc_ip(self, ip: str):
        self.rpc_ip_address_edit.setText(ip if ip else "")

    @property
    def rpc_port(self) -> int:
        try:
            return int(self.rpc_port_edit.text())
        except:
            return self.network_configs.configs[self.network.name].rpc_port

    @rpc_port.setter
    def rpc_port(self, port: int):
        self.rpc_port_edit.setText(str(port))

    @property
    def rpc_username(self) -> str:
        return self.rpc_username_edit.text()

    @rpc_username.setter
    def rpc_username(self, username: str):
        self.rpc_username_edit.setText(username if username else "")

    @property
    def rpc_password(self) -> str:
        return self.rpc_password_edit.text()

    @rpc_password.setter
    def rpc_password(self, password: str):
        self.rpc_password_edit.setText(password if password else "")



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

logger = logging.getLogger(__name__)

from typing import Callable, List

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QColor

from ...signals import UpdateFilter, UpdateFilterReason, WalletSignals
from .taglist import CustomListWidget, TagEditor, hash_color


class CategoryList(CustomListWidget):
    def __init__(
        self,
        categories: List[str],
        wallet_signals: WalletSignals,
        get_sub_texts: Callable[[], List[str]],
        parent=None,
        immediate_release=True,
    ) -> None:
        super().__init__(parent, enable_drag=False, immediate_release=immediate_release)

        self.categories = categories
        self.get_sub_texts = get_sub_texts
        self.wallet_signals = wallet_signals
        self.wallet_signals.updated.connect(self.refresh)
        self.refresh(UpdateFilter(refresh_all=True))

        # signals
        self.wallet_signals.updated.connect(self.refresh)
        self.wallet_signals.import_labels.connect(self.refresh)
        self.wallet_signals.import_bip329_labels.connect(self.refresh)
        self.wallet_signals.import_electrum_wallet_labels.connect(self.refresh)

        self.wallet_signals.language_switch.connect(self.refresh)

    def select_category(self, category: str):
        for i in range(self.count()):
            item = self.item(i)
            if item:
                item.setSelected(item.text() == category)

    def on_language_switch(self):
        self.refresh(UpdateFilter(refresh_all=True))

    @staticmethod
    def shoud_update(update_filter: UpdateFilter | None = None) -> bool:
        should_update = False
        if update_filter is None:
            return True
        if should_update or update_filter.refresh_all:
            should_update = True
        if should_update or update_filter.categories:
            should_update = True
        if should_update or update_filter.outpoints:
            should_update = True

        return should_update

    def refresh(self, update_filter: UpdateFilter | None = None) -> None:
        if not self.shoud_update(update_filter):
            return

        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")
        self.recreate(self.categories, sub_texts=self.get_sub_texts())

    @classmethod
    def color(cls, category) -> QColor:
        if not category:
            return QColor(255, 255, 255, 255)
        return hash_color(category)


class CategoryEditor(TagEditor):
    signal_category_added = pyqtSignal(str)

    def __init__(
        self,
        get_categories: Callable[[], List[str]],
        wallet_signals: WalletSignals,
        get_sub_texts: Callable[[], List[str]],
        parent=None,
        prevent_empty_categories=True,
    ) -> None:
        super().__init__(parent, get_categories(), sub_texts=get_sub_texts())

        self.get_categories = get_categories
        self.get_sub_texts = get_sub_texts
        self.wallet_signals = wallet_signals
        self.prevent_empty_categories = prevent_empty_categories

        self.updateUi()
        # signals
        self.wallet_signals.updated.connect(self.refresh)
        self.wallet_signals.import_labels.connect(self.refresh)
        self.wallet_signals.import_bip329_labels.connect(self.refresh)
        self.wallet_signals.import_electrum_wallet_labels.connect(self.refresh)

        self.list_widget.signal_tag_deleted.connect(self.on_delete)
        self.list_widget.signal_tag_added.connect(self.on_added)
        self.wallet_signals.language_switch.connect(self.updateUi)

    @classmethod
    def get_default_categories(cls) -> List[str]:
        return [cls.tr("KYC Exchange"), cls.tr("Private")]

    def add_default_categories(self):
        for category in self.get_default_categories():
            self.add(category)

    def updateUi(self) -> None:
        self.tag_name = self.tr("category")
        super().updateUi()
        self.refresh(UpdateFilter(refresh_all=True))

    def on_added(self, category) -> None:
        if not category or category in self.get_categories():
            return
        self.signal_category_added.emit(category)

        self.wallet_signals.updated.emit(
            UpdateFilter(categories=[category], reason=UpdateFilterReason.CategoryAdded)
        )

    def on_delete(self, category: str) -> None:
        if category not in self.get_categories():
            return
        self.wallet_signals.updated.emit(
            UpdateFilter(categories=[category], reason=UpdateFilterReason.CategoryDeleted)
        )

        if not self.get_categories() and self.prevent_empty_categories:
            self.list_widget.add("Default")
            self.wallet_signals.updated.emit(
                UpdateFilter(refresh_all=True, reason=UpdateFilterReason.CategoryDeleted)
            )

    def refresh(self, update_filter: UpdateFilter) -> None:
        if not CategoryList.shoud_update(update_filter):
            return

        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")
        self.list_widget.recreate(self.get_categories(), sub_texts=self.get_sub_texts())

    @classmethod
    def color(cls, category) -> QColor:
        if not category:
            return QColor(255, 255, 255, 255)
        return hash_color(category)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import logging.config
import logging.handlers
import platform
import sys
from pathlib import Path

import appdirs

from bitcoin_safe import __version__
from bitcoin_safe.logging_handlers import MailHandler, RelativePathFormatter


def setup_logging() -> None:

    # Configuring formatters
    relative_path_formatter = RelativePathFormatter(
        fmt="%(asctime)s - %(levelname)s - [%(threadName)s] - %(name)s - %(message)s"
    )

    # Configuring handlers
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(relative_path_formatter)

    app_name = "bitcoin_safe"
    config_dir = Path(appdirs.user_config_dir(app_name))
    config_dir.mkdir(parents=True, exist_ok=True)

    log_file = config_dir / ".bitcoin_safe.log"
    file_handler = logging.handlers.RotatingFileHandler(filename=log_file, maxBytes=1000000, backupCount=3)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(relative_path_formatter)

    custom_handler = (
        MailHandler()
    )  # Assuming MailHandler is correctly implemented in bitcoin_safe.logging_handlers
    custom_handler.setLevel(logging.CRITICAL)
    custom_handler.setFormatter(relative_path_formatter)
    # Assuming 'must_include_exc_info' is handled internally in the MailHandler implementation

    # Configuring loggers
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(console_handler)
    root_logger.addHandler(file_handler)
    root_logger.addHandler(custom_handler)
    root_logger.propagate = True

    logger = logging.getLogger(__name__)

    # Set the function to handle uncaught exceptions
    def handle_uncaught_exception(exc_type, exc_value, exc_traceback) -> None:
        logger.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = handle_uncaught_exception

    logger.info(f"========================= Starting Bitcoin Safe ========================")
    logger.info(f"Version: {__version__}")
    logger.info(f"Python version: {sys.version}. On platform: {describe_os_version()}")
    # logger.info(f"Logging to file: {str(logger.handlers[-1].filename)}")
    logger.info(f"Logging {logging.DEBUG} to {log_file}")


def describe_os_version() -> str:
    return platform.platform()


setup_logging()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import enum
import logging
from collections import defaultdict
from enum import Enum

from bitcoin_safe.pythonbdk_types import OutPoint

logger = logging.getLogger(__name__)
import threading
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
)

import bdkpython as bdk
from bitcoin_nostr_chat.signals_min import SignalsMin as NostrSignalsMin
from PyQt6.QtCore import pyqtSignal


class UpdateFilterReason(Enum):
    UserInput = enum.auto()
    UserImport = enum.auto()
    SourceLabelSyncer = enum.auto()
    Unknown = enum.auto()
    UserReplacedAddress = enum.auto()
    NewAddressRevealed = enum.auto()
    CategoryAssigned = enum.auto()
    CategoryAdded = enum.auto()
    CategoryRenamed = enum.auto()
    CategoryDeleted = enum.auto()
    GetUnusedCategoryAddress = enum.auto()
    RefreshCaches = enum.auto()
    CreatePSBT = enum.auto()
    TxCreator = enum.auto()
    TransactionChange = enum.auto()
    ForceRefresh = enum.auto()
    ChainHeightAdvanced = enum.auto()
    NewFxRates = enum.auto()


class UpdateFilter:
    def __init__(
        self,
        outpoints: Iterable[OutPoint] | None = None,
        addresses: Iterable[str] | None = None,
        categories: Iterable[Optional[str]] | None = None,
        txids: Iterable[str] | None = None,
        refresh_all=False,
        reason: UpdateFilterReason = UpdateFilterReason.Unknown,
    ) -> None:
        self.outpoints: Set[OutPoint] = set(outpoints) if outpoints else set()
        self.addresses: Set[str] = set(addresses) if addresses else set()
        self.categories = set(categories) if categories else set()
        self.txids = set(txids) if txids else set()
        self.refresh_all = refresh_all
        self.reason = reason

    def __key__(self) -> Tuple:
        return tuple(self.__dict__.items())

    def __str__(self) -> str:
        return str(self.__key__())

    def __hash__(self) -> int:
        return hash(str(self))


class SignalFunction:
    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name
        self.slots: Dict[str, Callable] = {}
        self.lock = threading.Lock()

    def connect(self, slot: Callable[[], Any], slot_name=None) -> None:
        with self.lock:
            key = slot_name if slot_name and (slot_name not in self.slots) else str(slot)
            self.slots[key] = slot

    def disconnect(self, slot) -> None:
        with self.lock:
            keys, values = list(self.slots.keys()), list(self.slots.values())
            if slot in values:
                idx = values.index(slot)
                del self.slots[keys[idx]]
            else:
                logger.debug(f"Tried to disconnect {slot}. But it is not in {values}. Skipping.")

    def __call__(self, *args, **kwargs) -> Dict[str, Any]:
        return self.emit(*args, **kwargs)

    def emit(self, *args, slot_name=None, **kwargs) -> Dict[str, Any]:
        allow_list = [slot_name] if isinstance(slot_name, str) else slot_name

        responses = {}
        if not self.slots:
            logger.debug(
                f"SignalFunction {self.name if self.name else ''}.emit() was called, but no listeners {self.slots} are listening."
            )

        delete_slots: List[Callable[[], Any]] = []
        with self.lock:
            for key, slot in self.slots.items():
                if allow_list and key not in allow_list:
                    continue

                name = f"{slot.__self__.__class__.__name__}." if hasattr(slot, "__self__") else ""
                name += f"{slot.__name__}{args, kwargs}"
                name += f" with key={key}" if key else ""
                try:
                    responses[key] = slot(*args, **kwargs)
                except:
                    logger.warning(
                        f"{slot} with key {key} caused an exception. {slot} with key {key} could not be called, perhaps because the object doesnt exisst anymore. The slot will be deleted."
                    )
                    delete_slots.append(slot)
                    continue
            logger.debug(
                f"SignalFunction {self.name if self.name else ''}.emit() --> Got {len(responses)} responses"
            )

        for slot in delete_slots:
            self.disconnect(slot)

        return responses


class SingularSignalFunction(SignalFunction):
    def connect(self, slot: Callable, slot_name=None) -> None:
        if not self.slots:
            super().connect(slot, slot_name=slot_name)
        else:
            raise Exception("Not allowed to add a second listener to this signal.")

    def emit(self, *args, **kwargs) -> Any:
        responses = super().emit(*args, **kwargs)
        return list(responses.values())[0] if responses else responses

    def __call__(self, *args, **kwargs) -> Any:
        return self.emit(*args, **kwargs)


class SignalsMin(NostrSignalsMin):
    pass


class WalletSignals(SignalsMin):
    updated = pyqtSignal(UpdateFilter)
    completions_updated = pyqtSignal()

    show_utxo = pyqtSignal(object)
    show_address = pyqtSignal(str, str)  # address, wallet_id

    export_bip329_labels = pyqtSignal(str)  # str= wallet_id
    export_labels = pyqtSignal(str)  # str= wallet_id
    import_labels = pyqtSignal(str)  # str= wallet_id
    import_bip329_labels = pyqtSignal(str)  # str= wallet_id
    import_electrum_wallet_labels = pyqtSignal(str)  # str= wallet_id

    finished_psbt_creation = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()
        self.get_display_balance = SignalFunction(name="get_display_balance")


class Signals(SignalsMin):
    """The idea here is to define events that might need to trigger updates of
    the UI or other events  (careful of circular loops)

    State what happended, NOT the intention:

        Good signal:
            utxos_changed
        And to this signal a listener might be triggered:
            utxo_list.update()

        A bad signal is:
            need_utxo_list_update


    I immediately break the rule however for pyqtSignal, which is a function call
    """

    open_tx_like = pyqtSignal(object)
    event_wallet_tab_closed = pyqtSignal()
    event_wallet_tab_added = pyqtSignal()

    chain_data_changed = pyqtSignal(str)  # the string is the reason
    notification = pyqtSignal(object)  # should be a Message instance

    show_network_settings = pyqtSignal()
    open_wallet = pyqtSignal(str)  # str= filepath
    finished_open_wallet = pyqtSignal(str)  # str= wallet_id
    create_qt_wallet_from_wallet = pyqtSignal(object, object, object)  # object = wallet, file_path, password
    close_qt_wallet = pyqtSignal(str)  # str = wallet_id

    request_manual_sync = pyqtSignal()
    signal_broadcast_tx = pyqtSignal(bdk.Transaction)

    # this is for non-wallet bound objects like UitxViewer
    any_wallet_updated = pyqtSignal(UpdateFilter)

    def __init__(self) -> None:
        super().__init__()
        self.get_wallets = SignalFunction(name="get_wallets")
        self.get_qt_wallets = SignalFunction(name="get_qt_wallets")
        self.get_network = SingularSignalFunction(name="get_network")
        self.get_mempool_url = SingularSignalFunction(name="get_mempool_url")

        self.wallet_signals: DefaultDict[str, WalletSignals] = defaultdict(WalletSignals)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import signal as syssignal
from collections import deque
from pathlib import Path

from bitcoin_usb.tool_gui import ToolGui

from bitcoin_safe import __version__
from bitcoin_safe.descriptors import MultipathDescriptor
from bitcoin_safe.gui.qt.about_dialog import LicenseDialog
from bitcoin_safe.gui.qt.category_list import CategoryEditor
from bitcoin_safe.gui.qt.descriptor_edit import DescriptorExport
from bitcoin_safe.gui.qt.descriptor_ui import KeyStoreUIs
from bitcoin_safe.gui.qt.language_chooser import LanguageChooser
from bitcoin_safe.gui.qt.notification_bar_regtest import NotificationBarRegtest
from bitcoin_safe.gui.qt.update_notification_bar import UpdateNotificationBar
from bitcoin_safe.gui.qt.wrappers import Menu, MenuBar
from bitcoin_safe.keystore import KeyStoreImporterTypes
from bitcoin_safe.pdfrecovery import make_and_open_pdf
from bitcoin_safe.threading_manager import ThreadingManager
from bitcoin_safe.util import rel_home_path_to_abs_path

logger = logging.getLogger(__name__)

import base64
import os
import sys
from typing import Deque, Dict, Iterable, List, Literal, Optional, Tuple, Union

import bdkpython as bdk
from bitcoin_qr_tools.bitcoin_video_widget import BitcoinVideoWidget
from bitcoin_qr_tools.data import Data, DataType
from PyQt6.QtCore import QCoreApplication, QProcess, Qt, QTimer
from PyQt6.QtGui import QCloseEvent, QIcon, QKeySequence, QShortcut
from PyQt6.QtWidgets import (
    QDialog,
    QFileDialog,
    QMainWindow,
    QScrollArea,
    QSizePolicy,
    QStyle,
    QSystemTrayIcon,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.search_tree_view import SearchWallets
from bitcoin_safe.gui.qt.wallet_steps import ImportXpubs, TutorialStep, WalletSteps

from ...config import UserConfig
from ...fx import FX
from ...mempool import MempoolData
from ...psbt_util import FeeInfo, SimplePSBT
from ...pythonbdk_types import get_outpoints
from ...signals import Signals
from ...storage import Storage
from ...tx import TxBuilderInfos, TxUiInfos, short_tx_id
from ...wallet import ProtoWallet, ToolsTxUiInfo, Wallet
from . import address_dialog
from .dialog_import import ImportDialog
from .dialogs import PasswordQuestion, WalletIdDialog, question_dialog
from .extended_tabwidget import ExtendedTabWidget, LoadingWalletTab
from .network_settings.main import NetworkSettingsUI
from .new_wallet_welcome_screen import NewWalletWelcomeScreen
from .qt_wallet import QTProtoWallet, QTWallet, QtWalletBase
from .ui_tx import UITx_Viewer, UITx_ViewerTab
from .util import (
    Message,
    MessageType,
    caught_exception_message,
    delayed_execution,
    read_QIcon,
    webopen,
)
from .utxo_list import UTXOList, UtxoListWithToolbar


class MainWindow(QMainWindow):
    def __init__(
        self,
        network: Literal["bitcoin", "regtest", "signet", "testnet"] | None = None,
        config: UserConfig | None = None,
        **kwargs,
    ) -> None:
        "If netowrk == None, then the network from the user config will be taken"
        super().__init__()
        config_present = UserConfig.exists() or config
        self.config = config if config else UserConfig.from_file()
        self.config.network = bdk.Network[network.upper()] if network else self.config.network
        self.new_startup_network: bdk.Network | None = None
        self.address_dialogs: Deque[address_dialog.AddressDialog] = deque(maxlen=1000)
        self._temp_bitcoin_video_widget: BitcoinVideoWidget | None = None
        self.setMinimumSize(600, 600)

        self.signals = Signals()
        self.qt_wallets: Dict[str, QTWallet] = {}
        self.threading_manager = ThreadingManager(threading_manager_name=self.__class__.__name__)

        self.fx = FX(threading_parent=self.threading_manager)
        self.language_chooser = LanguageChooser(self, self.config, [self.signals.language_switch])
        if not config_present:
            self.config.language_code = self.language_chooser.dialog_choose_language(self)
        self.language_chooser.set_language(self.config.language_code)
        self.hwi_tool_gui = ToolGui(self.config.network)
        self.setupUi(self)

        self.mempool_data = MempoolData(
            network_config=self.config.network_config,
            signals_min=self.signals,
            threading_parent=self.threading_manager,
        )
        self.mempool_data.set_data_from_mempoolspace()

        self.last_qtwallet: Optional[QTWallet] = None
        # connect the listeners
        self.signals.open_tx_like.connect(self.open_tx_like_in_tab)
        self.signals.get_network.connect(self.get_network)
        self.signals.get_mempool_url.connect(self.get_mempool_url)

        self.network_settings_ui = NetworkSettingsUI(
            self.config.network, self.config.network_configs, signals=self.signals
        )
        self.network_settings_ui.signal_apply_and_shutdown.connect(self.shutdown)
        self.signals.show_network_settings.connect(self.open_network_settings)

        self.welcome_screen = NewWalletWelcomeScreen(
            self.tab_wallets, network=self.config.network, signals=self.signals
        )

        # signals
        self.welcome_screen.signal_onclick_single_signature.connect(self.click_create_single_signature_wallet)
        self.welcome_screen.signal_onclick_multisig_signature.connect(
            self.click_create_multisig_signature_wallet
        )
        self.welcome_screen.signal_onclick_custom_signature.connect(self.click_custom_signature)
        self.signals.create_qt_wallet_from_wallet.connect(self.add_qt_wallet)
        self.signals.close_qt_wallet.connect(
            lambda wallet_id: self.remove_qt_wallet(self.qt_wallets.get(wallet_id))
        )

        self.signals.event_wallet_tab_added.connect(self.event_wallet_tab_added)
        self.signals.event_wallet_tab_closed.connect(self.event_wallet_tab_closed)
        self.signals.chain_data_changed.connect(self.sync)
        self.signals.request_manual_sync.connect(self.sync)
        self.signals.open_wallet.connect(self.open_wallet)
        self.signals.signal_broadcast_tx.connect(self.on_signal_broadcast_tx)
        self.signals.language_switch.connect(self.updateUI)

        self._init_tray()

        self.search_box = SearchWallets(
            get_qt_wallets=self.get_qt_wallets,
            signal_min=self.signals,
            parent=self.tab_wallets,
        )
        self.tab_wallets.set_top_right_widget(self.search_box)

        self.updateUI()
        self.setup_signal_handlers()

        delayed_execution(self.load_last_state, self)

    def get_qt_wallets(self) -> List[QTWallet]:
        return list(self.qt_wallets.values())

    def get_mempool_url(self) -> str:
        return self.config.network_config.mempool_url

    def get_network(self) -> bdk.Network:
        return self.config.network

    def load_last_state(self) -> None:

        opened_qt_wallets = self.open_last_opened_wallets()
        if not opened_qt_wallets:
            self.welcome_screen.add_new_wallet_welcome_tab()

        self.open_last_opened_tx()

    def set_title(self) -> None:
        title = "Bitcoin Safe"
        if self.config.network != bdk.Network.BITCOIN:
            title += f" - {self.config.network.name}"
        if qt_wallet := self.get_qt_wallet():
            title += f" - {qt_wallet.wallet.id}"
        self.setWindowTitle(title)

    def setupUi(self, MainWindow: QWidget) -> None:
        logger.debug(f"start setupUi")
        # sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        # sizePolicy.setHorizontalStretch(0)
        # sizePolicy.setVerticalStretch(0)
        # sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        # MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setWindowIcon(read_QIcon("logo.svg"))
        w, h = 900, 600
        MainWindow.resize(w, h)
        MainWindow.setMinimumSize(w, h)

        #####
        self.tab_wallets = ExtendedTabWidget(object, parent=self)
        self.tab_wallets.tabBar().setExpanding(True)  # type: ignore[union-attr]  # This will expand tabs to fill the tab widget width
        self.tab_wallets.setTabBarAutoHide(False)
        self.tab_wallets.setMovable(True)  # Enable tab reordering
        self.tab_wallets.setTabsClosable(True)
        self.tab_wallets.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.tab_wallets.signal_tab_bar_visibility.connect(self.updateUI)
        # Connect signals to slots
        self.tab_wallets.tabCloseRequested.connect(self.close_tab)
        self.tab_wallets.currentChanged.connect(self.set_title)

        # central_widget
        central_widget = QScrollArea()
        vbox = QVBoxLayout(central_widget)
        vbox.setSpacing(0)
        vbox.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins
        # header bar about testnet coins
        if self.config.network != bdk.Network.BITCOIN:
            notification_bar = NotificationBarRegtest(
                open_network_settings=self.open_network_settings,
                network=self.config.network,
                signals_min=self.signals,
            )
            vbox.addWidget(notification_bar)

        self.update_notification_bar = UpdateNotificationBar(
            signals_min=self.signals, threading_parent=self.threading_manager
        )
        self.update_notification_bar.check()  # TODO: disable this, after it got more stable
        vbox.addWidget(self.update_notification_bar)

        vbox.addWidget(self.tab_wallets)
        self.setCentralWidget(central_widget)

        self.setMinimumWidth(640)
        self.setMinimumHeight(400)
        if self.config.is_maximized:
            self.showMaximized()

        # self.setWindowIcon(read_QIcon("electrum.png"))
        self.init_menubar()
        self.set_title()
        logger.debug(f"done setupUi")

    def init_menubar(self) -> None:
        self.menubar = MenuBar()
        # menu wallet
        self.menu_wallet = self.menubar.add_menu("")
        self.menu_action_new_wallet = self.menu_wallet.add_action("", self.new_wallet)
        self.menu_action_new_wallet.setIcon(
            (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_FileDialogNewFolder)
        )

        self.menu_action_open_wallet = self.menu_wallet.add_action("", self.open_wallet)
        self.menu_action_open_wallet.setShortcut(QKeySequence("CTRL+O"))
        self.menu_action_open_wallet.setIcon(
            (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon)
        )

        self.menu_wallet_recent = self.menu_wallet.add_menu("")

        self.menu_action_save_current_wallet = self.menu_wallet.add_action("", self.save_qt_wallet)
        self.menu_action_save_current_wallet.setShortcut(QKeySequence("CTRL+S"))
        self.menu_action_save_current_wallet.setIcon(
            (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_DialogSaveButton)
        )
        self.menu_wallet.addSeparator()

        self.menu_action_search = self.menu_wallet.add_action("", self.focus_search_box)
        self.menu_action_search.setShortcut(QKeySequence("CTRL+F"))
        self.menu_action_search.setIcon(read_QIcon("search.svg"))

        # change wallet
        self.menu_wallet_change = self.menu_wallet.add_menu("")
        self.menu_wallet_change.setIcon(read_QIcon("password.svg"))
        self.menu_action_rename_wallet = self.menu_wallet_change.add_action("", self.change_wallet_id)
        self.menu_action_change_password = self.menu_wallet_change.add_action("", self.change_wallet_password)
        self.menu_action_change_password.setIcon(read_QIcon("password.svg"))

        # export wallet
        self.menu_wallet_export = self.menu_wallet.add_menu("")
        self.menu_action_export_pdf = self.menu_wallet_export.add_action(
            "", self.export_wallet_pdf, icon=read_QIcon("descriptor-backup.svg")
        )
        self.menu_action_export_for_coldcard = self.menu_wallet_export.add_action(
            "", self.export_wallet_for_coldcard, icon=read_QIcon("coldcard-only.svg")
        )
        self.menu_action_export_descriptor = self.menu_wallet_export.add_action(
            "", self.export_wallet_for_coldcard_q
        )

        self.menu_wallet.addSeparator()
        self.menu_action_refresh_wallet = self.menu_wallet.add_action(
            "", self.signals.request_manual_sync.emit
        )
        self.menu_action_refresh_wallet.setShortcut(QKeySequence("F5"))
        self.menu_action_refresh_wallet.setIcon(
            (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_BrowserReload)
        )

        # menu transaction
        self.menu_tools = self.menubar.add_menu("")

        self.menu_action_open_hwi_manager = self.menu_tools.add_action(
            "",
            self.hwi_tool_gui.show,
            icon=read_QIcon(KeyStoreImporterTypes.hwi.icon_filename),
        )

        self.menu_load_transaction = self.menu_tools.add_menu("")
        self.menu_action_open_tx_file = self.menu_load_transaction.add_action(
            "",
            self.open_tx_file,
            icon=(self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
        )
        self.menu_action_open_tx_from_str = self.menu_load_transaction.add_action(
            "",
            self.dialog_open_tx_from_str,
            icon=(self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_FileIcon),
        )
        self.menu_action_open_tx_from_str.setShortcut(QKeySequence("CTRL+L"))

        self.menu_action_load_tx_from_qr = self.menu_load_transaction.add_action(
            "", self.load_tx_like_from_qr, icon=read_QIcon("qr-code.svg")
        )

        # menu settings
        self.menu_settings = self.menubar.add_menu("")
        self.menu_action_network_settings = self.menu_settings.add_action(
            "",
            self.open_network_settings,
            icon=(self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_DriveNetIcon),
        )
        self.menu_action_network_settings.setShortcut(QKeySequence("CTRL+P"))
        self.menu_action_toggle_tutorial = self.menu_settings.add_action("", self.toggle_tutorial)
        self.language_menu = self.menu_settings.add_menu("")

        # menu about
        self.menu_about = self.menubar.add_menu("")
        self.menu_action_version = self.menu_about.add_action(
            "", lambda: webopen("https://github.com/andreasgriffin/bitcoin-safe/releases")
        )
        self.menu_action_check_update = self.menu_about.add_action(
            "", self.update_notification_bar.check_and_make_visible
        )
        self.menu_action_check_update.setShortcut(QKeySequence("CTRL+U"))

        def menu_action_license():
            LicenseDialog().exec()

        self.menu_action_license = self.menu_about.add_action("", menu_action_license)

        # assigning menu bar
        self.setMenuBar(self.menubar)

        # Populate recent wallets menu
        self.populate_recent_wallets_menu()

        # other shortcuts
        self.shortcut_close_tab = QShortcut(QKeySequence("Ctrl+W"), self)
        self.shortcut_close_tab.activated.connect(lambda: self.close_tab(self.tab_wallets.currentIndex()))

    def showEvent(self, event) -> None:
        super().showEvent(event)
        # self.updateUI()

    def updateUI(self) -> None:

        # menu
        self.menu_wallet.setTitle(self.tr("&Wallet"))
        self.menu_action_new_wallet.setText(self.tr("&New Wallet"))
        self.menu_action_open_wallet.setText(self.tr("&Open Wallet"))
        self.menu_wallet_recent.setTitle(self.tr("Open &Recent"))
        self.menu_action_save_current_wallet.setText(self.tr("&Save Current Wallet"))
        self.menu_action_search.setText(self.tr("&Search"))
        self.menu_wallet_change.setTitle(self.tr("&Change"))
        self.menu_wallet_export.setTitle(self.tr("&Export"))
        self.menu_action_rename_wallet.setText(self.tr("&Rename Wallet"))
        self.menu_action_change_password.setText(self.tr("&Change Password"))
        self.menu_action_export_for_coldcard.setText(self.tr("&Export Coldcard txt file"))
        self.menu_action_export_pdf.setText(self.tr("&Export Wallet PDF"))
        self.menu_action_export_descriptor.setText(self.tr("&Export Descriptor"))
        self.menu_action_refresh_wallet.setText(self.tr("Re&fresh"))
        self.menu_tools.setTitle(self.tr("&Tools"))
        self.menu_action_open_hwi_manager.setText(self.tr("&USB Signer Tools"))
        self.menu_load_transaction.setTitle(self.tr("&Load Transaction or PSBT"))
        self.menu_action_open_tx_file.setText(self.tr("From &file"))
        self.menu_action_open_tx_from_str.setText(self.tr("From &text"))
        self.menu_action_load_tx_from_qr.setText(self.tr("From &QR Code"))
        self.menu_settings.setTitle(self.tr("&Settings"))
        self.menu_action_network_settings.setText(self.tr("&Network Settings"))
        self.menu_action_toggle_tutorial.setText(self.tr("&Show/Hide Tutorial"))
        languages = "&Languages"
        local_languages = self.tr("&Languages")
        if local_languages != languages:
            languages += f" - {local_languages}"
        self.language_menu.setTitle(languages)
        self.language_chooser.populate_language_menu(self.language_menu)
        self.menu_about.setTitle(self.tr("&About"))
        self.menu_action_version.setText(self.tr("&Version: {}").format(__version__))
        self.menu_action_check_update.setText(self.tr("&Check for update"))
        self.menu_action_license.setText(self.tr("&License"))

        # the search fields
        for qt_wallet in self.qt_wallets.values():
            if self.tab_wallets.top_right_widget:
                main_search_field_hidden = (
                    self.tab_wallets.count() <= 1
                ) and self.tab_wallets.tabBarAutoHide()
                self.tab_wallets.top_right_widget.setVisible(not main_search_field_hidden)
                if qt_wallet.tabs.top_right_widget:
                    qt_wallet.tabs.top_right_widget.setVisible(main_search_field_hidden)

    def focus_search_box(self):
        self.search_box.search_field.setFocus(Qt.FocusReason.ShortcutFocusReason)

    def populate_recent_wallets_menu(self) -> None:
        self.menu_wallet_recent.clear()

        def factory(filepath):
            def f(*args):
                self.open_wallet(file_path=filepath)

            return f

        for filepath in reversed(self.config.recently_open_wallets[self.config.network]):
            if not Path(filepath).exists():
                continue
            self.menu_wallet_recent.add_action(os.path.basename(filepath), factory(filepath=filepath))

    def change_wallet_id(self) -> Optional[str]:
        qt_wallet = self.get_qt_wallet()
        if not qt_wallet:
            Message(self.tr("Please select the wallet"))
            return None

        old_id = qt_wallet.wallet.id

        # ask for wallet name
        dialog = WalletIdDialog(Path(self.config.wallet_dir), prefilled=old_id)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            new_wallet_id = dialog.wallet_id
            new_wallet_filename = dialog.filename
            logger.info(f"new wallet name: {new_wallet_id}")
        else:
            return None

        # in the wallet
        qt_wallet.wallet.set_wallet_id(new_wallet_id)
        # change dict key
        self.qt_wallets[new_wallet_id] = qt_wallet
        del self.qt_wallets[old_id]

        # tab text
        self.tab_wallets.setTabText(self.tab_wallets.indexOf(qt_wallet.tab), new_wallet_id)

        # save under new filename
        old_filepath = qt_wallet.file_path
        directory, old_filename = os.path.split(old_filepath)

        new_file_path = os.path.join(directory, new_wallet_filename)

        qt_wallet.move_wallet_file(new_file_path)
        self.save_qt_wallet(qt_wallet)
        logger.info(f"Saved {old_filepath} under new name {qt_wallet.file_path}")
        self.set_title()
        return new_wallet_id

    def change_wallet_password(self) -> None:
        qt_wallet = self.get_qt_wallet()
        if not qt_wallet:
            Message(self.tr("Please select the wallet"))
            return

        qt_wallet.change_password()

    def on_signal_broadcast_tx(self, transaction: bdk.Transaction) -> None:
        def f_sync_all(qt_wallets: List[QTWallet]):
            for qt_wallet in qt_wallets:
                qt_wallet.sync()

        qt_wallets_to_sync: List[QTWallet] = []

        last_qt_wallet_involved: Optional[QTWallet] = None
        for qt_wallet in self.qt_wallets.values():
            if qt_wallet.wallet.transaction_related_to_my_addresses(transaction):
                qt_wallets_to_sync.append(qt_wallet)
                last_qt_wallet_involved = qt_wallet

        if last_qt_wallet_involved:
            self.tab_wallets.setCurrentWidget(last_qt_wallet_involved.tab)
            last_qt_wallet_involved.tabs.setCurrentWidget(last_qt_wallet_involved.history_tab)

        QTimer.singleShot(500, lambda: f_sync_all(qt_wallets_to_sync))

    def on_tab_changed(self, index: int) -> None:
        qt_wallet = self.get_qt_wallet(self.tab_wallets.widget(index))
        if qt_wallet:
            self.last_qtwallet = qt_wallet

    def _init_tray(self) -> None:
        self.tray = QSystemTrayIcon(read_QIcon("logo.svg"), self)
        self.tray.setToolTip("Bitcoin Safe")

        menu = Menu(self)
        menu.add_action(text="&Exit", slot=self.close)

        self.tray.setContextMenu(menu)
        self.tray.activated.connect(self.onTrayIconActivated)

        self.signals.notification.connect(self.show_message_as_tray_notification)
        self.tray.show()

    def show_message_as_tray_notification(self, message: Message) -> None:
        icon, _ = message.get_icon_and_title()
        title = message.title or "Bitcoin Safe"
        if message.msecs:
            self.tray.showMessage(title, message.msg, Message.system_tray_icon(icon), message.msecs)
            return
        self.tray.showMessage(title, message.msg, Message.system_tray_icon(icon))

    def onTrayIconActivated(self, reason: QSystemTrayIcon.ActivationReason) -> None:
        # Attempts to bring the window to the foreground
        self.raise_()
        self.activateWindow()
        # if reason == QSystemTrayIcon.ActivationReason.Trigger:
        #     Message(self.tr("test"), no_show=True).emit_with(self.signals.notification)

    def open_network_settings(self) -> None:
        self.network_settings_ui.exec()

    def export_wallet_for_coldcard(self, wallet: Wallet | None = None) -> None:
        qt_wallet = self.get_qt_wallet(if_none_serve_last_active=True)
        if not qt_wallet or not qt_wallet.wallet:
            Message(self.tr("Please select the wallet first."), type=MessageType.Warning)
            return

        qt_wallet.export_wallet_for_coldcard()

    def export_wallet_for_coldcard_q(self, wallet: Wallet | None = None) -> None:
        qt_wallet = self.get_qt_wallet(if_none_serve_last_active=True)
        if not qt_wallet or not qt_wallet.wallet:
            Message(self.tr("Please select the wallet first."), type=MessageType.Warning)
            return

        edit = qt_wallet.wallet_descriptor_ui.edit_descriptor
        dialog = DescriptorExport(
            MultipathDescriptor.from_descriptor_str(edit.text(), qt_wallet.wallet.network),
            qt_wallet.signals,
            parent=self,
            network=self.config.network,
            threading_parent=self.threading_manager,
        )
        dialog.show()

    def export_wallet_pdf(self, wallet: Wallet | None = None) -> None:
        qt_wallet = self.get_qt_wallet(if_none_serve_last_active=True)
        if not qt_wallet or not qt_wallet.wallet:
            Message(self.tr("Please select the wallet first."), type=MessageType.Warning)
            return

        make_and_open_pdf(qt_wallet.wallet, lang_code=self.language_chooser.get_current_lang_code())

    def open_tx_file(self, file_path: Optional[str] = None) -> None:
        if not file_path:
            file_path, _ = QFileDialog.getOpenFileName(
                self,
                self.tr("Open Transaction/PSBT"),
                "",
                self.tr("All Files (*);;PSBT (*.psbt);;Transation (*.tx)"),
            )
            if not file_path:
                logger.info("No file selected")
                return

        logger.info(self.tr("Selected file: {file_path}").format(file_path=file_path))
        with open(file_path, "rb") as file:
            string_content = file.read()

        self.signals.open_tx_like.emit(string_content)

    def fetch_txdetails(self, txid: str) -> Optional[bdk.TransactionDetails]:
        for qt_wallet in self.qt_wallets.values():
            tx_details = qt_wallet.wallet.get_tx(txid)
            if tx_details:
                return tx_details
        return None

    def open_tx_like_in_tab(
        self,
        txlike: Union[
            bdk.TransactionDetails,
            bdk.Transaction,
            bdk.PartiallySignedTransaction,
            TxBuilderInfos,
            TxUiInfos,
            bytes,
            str,
        ],
    ) -> None:
        logger.info(f"Trying to open tx with type {type(txlike)}")

        # first do the bdk instance cases
        if isinstance(txlike, (bdk.TransactionDetails, bdk.Transaction)):
            self.open_tx_in_tab(txlike)
            return None

        if isinstance(txlike, (bdk.PartiallySignedTransaction, TxBuilderInfos)):
            self.open_psbt_in_tab(txlike)
            return None

        if isinstance(txlike, TxUiInfos):
            wallet = ToolsTxUiInfo.get_likely_source_wallet(txlike, self.signals)

            if not wallet:
                logger.info(
                    f"Could not identify the wallet belonging to the transaction inputs. Trying to open anyway..."
                )
                current_qt_wallet = self.get_qt_wallet(if_none_serve_last_active=True)
                wallet = current_qt_wallet.wallet if current_qt_wallet else None
            if not wallet:
                Message(self.tr("No wallet open. Please open the sender wallet to edit this thransaction."))
                return None

            qt_wallet = self.qt_wallets.get(wallet.id)
            if not qt_wallet:
                Message(self.tr(" Please open the sender wallet to edit this thransaction."))
                return None
            self.tab_wallets.setCurrentWidget(qt_wallet.tab)
            qt_wallet.tabs.setCurrentWidget(qt_wallet.send_tab)

            ToolsTxUiInfo.pop_change_recipient(txlike, wallet)

            qt_wallet.uitx_creator.set_ui(txlike)
            return None

        # try to convert a bytes like object to a string
        if isinstance(txlike, bytes):
            if txlike[:5] == b"psbt\xff":
                # convert a psbt in the default base64 encoding
                txlike = base64.encodebytes(txlike).decode()
                txlike = txlike.replace("\n", "").strip()
            else:
                # try to convert to str
                txlike = str(txlike)

        if isinstance(txlike, str):
            try:
                res = Data.from_str(txlike, self.config.network)
            except:
                Message(self.tr("Could not decode this string"), type=MessageType.Error)
                return None
            if res.data_type == DataType.Txid:
                txdetails = self.fetch_txdetails(res.data)
                if txdetails:
                    self.open_tx_in_tab(txdetails)
                    return None
                if not txlike:
                    raise Exception(f"txid {res.data} could not be found in wallets")
            elif res.data_type == DataType.PSBT:
                self.open_psbt_in_tab(res.data)
                return None
            elif res.data_type == DataType.Tx:
                self.open_tx_in_tab(res.data)
                return None
            else:
                logger.warning(f"DataType {res.data_type.name} was not handled.")
        return None

    def load_tx_like_from_qr(self) -> None:
        def result_callback(data: Data) -> None:
            if data.data_type in [
                DataType.PSBT,
                DataType.Tx,
                DataType.Txid,
            ]:
                self.open_tx_like_in_tab(data.data)

        if self._temp_bitcoin_video_widget:
            self._temp_bitcoin_video_widget.close()
        self._temp_bitcoin_video_widget = BitcoinVideoWidget()
        self._temp_bitcoin_video_widget.signal_data.connect(result_callback)
        self._temp_bitcoin_video_widget.show()
        return None

    def dialog_open_tx_from_str(self) -> None:
        def process_input(s: str) -> None:
            self.open_tx_like_in_tab(s)

        tx_dialog = ImportDialog(
            network=self.config.network,
            on_open=process_input,
            window_title=self.tr("Open Transaction or PSBT"),
            text_button_ok=self.tr("OK"),
            text_instruction_label=self.tr(
                "Please paste your Bitcoin Transaction or PSBT in here, or drop a file"
            ),
            text_placeholder=self.tr("Paste your Bitcoin Transaction or PSBT in here or drop a file"),
        )
        tx_dialog.show()

    def get_tab_with_title(self, tab_widget: QTabWidget, title) -> Optional[int]:
        for i in range(tab_widget.count()):
            if title == tab_widget.tabText(i):
                return i
        return None

    def open_tx_in_tab(
        self, txlike: Union[bdk.Transaction, bdk.TransactionDetails]
    ) -> Optional[Tuple[UITx_ViewerTab, UITx_Viewer]]:
        tx: bdk.Transaction | None = None
        fee = None
        confirmation_time = None

        if isinstance(txlike, bdk.Transaction):
            # try to get all details from wallets
            tx_details = self.fetch_txdetails(txlike.txid())
            if tx_details:
                txlike = tx_details

        if isinstance(txlike, bdk.TransactionDetails):
            logger.debug(f"Got a PartiallySignedTransaction")
            tx = txlike.transaction
            fee = txlike.fee
            if fee is None and txlike.transaction.is_coin_base():
                fee = 0
            confirmation_time = txlike.confirmation_time
        elif isinstance(txlike, bdk.Transaction):
            tx = txlike

        if not tx:
            logger.error(f"could not open {tx}")
            return None

        title = self.tr("Transaction {txid}").format(txid=short_tx_id(tx.txid()))
        data = Data.from_tx(tx)

        # check if the same tab with exactly the same data is open already
        tab_idx = self.get_tab_with_title(self.tab_wallets, title)
        if tab_idx is not None and isinstance(tab_data := self.tab_wallets.tabData(tab_idx), UITx_Viewer):
            # if the tab_data is a tx, then just dismiss the tx
            if tab_data.data.data_type == DataType.Tx:
                self.tab_wallets.setCurrentIndex(tab_idx)
                return None
            # if tab_data is a psbt, then add the signature from tx
            if tab_data.data.data_type == DataType.PSBT:
                tab_data.tx_received(tx)
                self.tab_wallets.setCurrentIndex(tab_idx)
                return None

        utxo_list = UTXOList(
            self.config,
            self.signals,
            get_outpoints=lambda: get_outpoints(tx),
            hidden_columns=[
                UTXOList.Columns.OUTPOINT,
                UTXOList.Columns.PARENTS,
            ],
            keep_outpoint_order=True,
            # the ADDRESS. ROLE SORT ORDER saves the order of the get_outpoints
            sort_column=UTXOList.Columns.ADDRESS,
            sort_order=Qt.SortOrder.AscendingOrder,
        )

        widget_utxo_with_toolbar = UtxoListWithToolbar(utxo_list, self.config, self.tab_wallets)

        viewer = UITx_Viewer(
            self.config,
            self.signals,
            self.fx,
            widget_utxo_with_toolbar,
            network=self.config.network,
            mempool_data=self.mempool_data,
            fee_info=FeeInfo(fee, tx.vsize(), is_estimated=False) if fee is not None else None,
            confirmation_time=confirmation_time,
            blockchain=self.get_blockchain_of_any_wallet(),
            data=data,
            parent=self,
            threading_parent=self.threading_manager,
        )

        self.tab_wallets.add_tab(
            tab=viewer,
            icon=read_QIcon("send.svg"),
            description=title,
            focus=True,
            data=viewer,
        )

        return viewer, viewer

    def open_psbt_in_tab(
        self,
        tx: Union[
            bdk.PartiallySignedTransaction, TxBuilderInfos, bdk.TxBuilderResult, str, bdk.TransactionDetails
        ],
    ) -> Optional[Tuple[UITx_ViewerTab, UITx_Viewer]]:
        psbt: bdk.PartiallySignedTransaction | None = None
        fee_info: Optional[FeeInfo] = None

        logger.debug(f"tx is of type {type(tx)}")

        # converting to TxBuilderResult
        if isinstance(tx, TxBuilderInfos):
            tx = tx.builder_result  # then it is processed in the next if stament
            logger.debug(f"Converted TxBuilderInfos --> {type(tx)}")

        if isinstance(tx, bdk.TxBuilderResult):
            psbt = tx.psbt
            fee_info = FeeInfo.estimate_segwit_fee_rate_from_psbt(psbt)
            logger.debug(f"Converted TxBuilderResult --> {type(psbt)}")

        if isinstance(tx, bdk.PartiallySignedTransaction):
            logger.debug(f"Got a PartiallySignedTransaction")
            psbt = tx
            fee_info = FeeInfo.estimate_segwit_fee_rate_from_psbt(psbt)

        if isinstance(tx, str):
            psbt = bdk.PartiallySignedTransaction(tx)
            logger.debug(f"Converted str to {type(tx)}")
            fee_info = FeeInfo.estimate_segwit_fee_rate_from_psbt(psbt)

        if isinstance(tx, bdk.TransactionDetails):
            logger.debug("is bdk.TransactionDetails")
            raise Exception("cannot handle TransactionDetails")

        if not psbt:
            logger.error(f"{tx} could not be converted to a psbt")
            return None

        data = Data.from_psbt(psbt)
        title = self.tr("PSBT {txid}").format(txid=short_tx_id(psbt.txid()))

        # check if the same tab with exactly the same data is open already
        tab_idx = self.get_tab_with_title(self.tab_wallets, title)
        if tab_idx is not None and isinstance(tab_data := self.tab_wallets.tabData(tab_idx), UITx_Viewer):
            # if the tab_data is a tx, then just dismiss the psbt (a tx is better than a psbt)
            if tab_data.data.data_type == DataType.Tx:
                self.tab_wallets.setCurrentIndex(tab_idx)
                return None
            # if tab_data is a psbt, then add the signature from data
            if tab_data.data.data_type == DataType.PSBT:
                tab_data.signature_added(psbt)
                self.tab_wallets.setCurrentIndex(tab_idx)
                return None

        utxo_list = UTXOList(
            self.config,
            self.signals,
            get_outpoints=lambda: get_outpoints(psbt.extract_tx()),
            hidden_columns=[
                UTXOList.Columns.OUTPOINT,
                UTXOList.Columns.PARENTS,
            ],
            txout_dict=SimplePSBT.from_psbt(psbt).get_prev_txouts(),
            keep_outpoint_order=True,
            # the ADDRESS. ROLE SORT ORDER saves the order of the get_outpoints
            sort_column=UTXOList.Columns.ADDRESS,
            sort_order=Qt.SortOrder.AscendingOrder,
        )

        widget_utxo_with_toolbar = UtxoListWithToolbar(utxo_list, self.config, parent=self.tab_wallets)

        viewer = UITx_Viewer(
            self.config,
            self.signals,
            self.fx,
            widget_utxo_with_toolbar,
            network=self.config.network,
            mempool_data=self.mempool_data,
            fee_info=fee_info,
            blockchain=self.get_blockchain_of_any_wallet(),
            data=data,
            parent=self,
            threading_parent=self.threading_manager,
        )

        self.tab_wallets.add_tab(
            tab=viewer,
            icon=read_QIcon("qr-code.svg"),
            description=title,
            focus=True,
            data=viewer,
        )

        return viewer, viewer

    def open_last_opened_wallets(self) -> List[QTWallet]:
        opened_wallets: List[QTWallet] = []
        for file_path in self.config.last_wallet_files.get(str(self.config.network), []):
            qt_wallet = self.open_wallet(file_path=str(rel_home_path_to_abs_path(file_path)))
            if qt_wallet:
                opened_wallets.append(qt_wallet)
        return opened_wallets

    def open_last_opened_tx(self) -> None:
        for serialized in self.config.opened_txlike.get(str(self.config.network), []):
            self.open_tx_like_in_tab(serialized)

    # def advance_tips_in_background(self, wallet: Wallet):
    #     def do():
    #         return wallet.get_address()

    #     def on_error(packed_error_info):
    #         logger.error("Exception in advance_tips_in_background")

    #     def on_done(data):
    #         pass

    #     def on_success(data):
    #         pass

    #     TaskThread( signals_min=self.signals).add_and_start(do, on_success, on_done, on_error)

    def open_wallet(self, file_path: Optional[str] = None) -> Optional[QTWallet]:
        if not file_path:
            file_path, _ = QFileDialog.getOpenFileName(
                self,
                self.tr("Open Wallet"),
                self.config.wallet_dir,
                self.tr("Wallet Files (*.wallet);;All Files (*)"),
            )
            if not file_path:
                logger.info("No file selected")
                return None

        # make sure this wallet isn't open already by this instance
        opened_file_paths = [qt_wallet.file_path for qt_wallet in self.qt_wallets.values()]
        if file_path in opened_file_paths:
            Message(self.tr("The wallet {file_path} is already open.").format(file_path=file_path))
            return None

        if not QTWallet.get_wallet_lockfile(Path(file_path)):
            if not question_dialog(
                self.tr(
                    "The wallet {file_path} is already open.  Do you want to open the wallet anyway?"
                ).format(file_path=file_path),
                title=self.tr("Wallet already open"),
            ):
                return None

        logger.info(f"Selected file: {file_path}")
        if not os.path.isfile(file_path):
            Message(
                self.tr("There is no such file: {file_path}").format(file_path=file_path),
                type=MessageType.Error,
            )
            return None
        password = None
        if Storage().has_password(file_path):
            direcory, filename = os.path.split(file_path)
            ui_password_question = PasswordQuestion(
                label_text=self.tr("Please enter the password for {filename}:").format(filename=filename)
            )
            password = ui_password_question.ask_for_password()
        try:
            wallet: Wallet = Wallet.from_file(file_path, self.config, password)
        except Exception as e:
            # the file could also be corrupted, but the "wrong password" is by far the likliest
            caught_exception_message(e, "Wrong password. Wallet could not be loaded.")
            QTWallet.remove_lockfile(Path(file_path))
            return None
        # self.advance_tips_in_background(wallet)

        if wallet.id in self.qt_wallets:
            Message(
                self.tr("A wallet with id {name} is already open. Please close it first.").format(
                    name=wallet.id
                )
            )
            return None

        qt_wallet = self.add_qt_wallet(wallet)
        qt_wallet.password = password
        qt_wallet.file_path = file_path
        qt_wallet.sync()

        self.add_recently_open_wallet(qt_wallet.file_path)

        self.signals.finished_open_wallet.emit(wallet.id)
        return qt_wallet

    def save_qt_wallet(self, qt_wallet: QTWallet | None = None) -> None:
        qt_wallet = qt_wallet if qt_wallet else self.get_qt_wallet()
        if qt_wallet:
            qt_wallet.save()
            self.add_recently_open_wallet(qt_wallet.file_path)

    def save_all_wallets(self) -> None:
        for qt_wallet in self.qt_wallets.values():
            self.save_qt_wallet(qt_wallet=qt_wallet)

    def write_current_open_txs_to_config(self) -> None:
        l = []

        for index in range(self.tab_wallets.count()):
            # Get the widget for the current tab
            tab = self.tab_wallets.widget(index)
            if isinstance(tab, UITx_ViewerTab):
                l.append(tab.serialize())

        self.config.opened_txlike[str(self.config.network)] = l

    def click_create_single_signature_wallet(self) -> None:
        qtprotowallet = self.create_qtprotowallet((1, 1), show_tutorial=True)
        if qtprotowallet:
            qtprotowallet.wallet_descriptor_ui.disable_fields()

    def click_create_multisig_signature_wallet(self) -> None:
        qtprotowallet = self.create_qtprotowallet((2, 3), show_tutorial=True)
        if qtprotowallet:
            qtprotowallet.wallet_descriptor_ui.disable_fields()

    def click_custom_signature(self) -> None:
        qtprotowallet = self.create_qtprotowallet((3, 5), show_tutorial=False)

    def new_wallet(self) -> None:
        self.welcome_screen.add_new_wallet_welcome_tab()

    def new_wallet_id(self) -> str:
        return f'{self.tr("new")}{len(self.qt_wallets)}'

    def create_qtwallet_from_protowallet(self, protowallet: ProtoWallet) -> QTWallet:
        wallet = Wallet.from_protowallet(
            protowallet, self.config, default_category=CategoryEditor.get_default_categories()[0]
        )

        qt_wallet = self.add_qt_wallet(wallet)
        # adding these should only be done at wallet creation
        qt_wallet.address_list_tags.add_default_categories()
        self.save_qt_wallet(qt_wallet)
        qt_wallet.sync()
        return qt_wallet

    def create_qtwallet_from_ui(
        self,
        wallet_tab: QWidget,
        protowallet: ProtoWallet,
        keystore_uis: KeyStoreUIs,
    ) -> None:
        try:
            if keystore_uis.ask_accept_unexpected_origins():
                qt_wallet = self.create_qtwallet_from_protowallet(protowallet=protowallet)
                self.tab_wallets.removeTab(self.tab_wallets.indexOf(wallet_tab))
                qt_wallet.tabs.setCurrentWidget(qt_wallet.history_tab)

            else:
                return
        except Exception as e:
            Message(str(e), type=MessageType.Error)

    def create_qtwallet_from_qtprotowallet(self, qtprotowallet: QTProtoWallet) -> None:
        self.create_qtwallet_from_ui(
            wallet_tab=qtprotowallet.tab,
            protowallet=qtprotowallet.protowallet,
            keystore_uis=qtprotowallet.wallet_descriptor_ui.keystore_uis,
        )

    def create_qtprotowallet(
        self, m_of_n: Tuple[int, int], show_tutorial: bool = False
    ) -> Optional[QTProtoWallet]:

        # ask for wallet name
        dialog = WalletIdDialog(Path(self.config.wallet_dir))
        if dialog.exec() == QDialog.DialogCode.Accepted:
            wallet_id = dialog.wallet_id
            logger.info(f"new wallet name: {wallet_id}")
        else:
            return None

        m, n = m_of_n
        protowallet = ProtoWallet(
            threshold=m,
            keystores=[None for i in range(n)],
            network=self.config.network,
            wallet_id=wallet_id,
        )

        qtprotowallet = QTProtoWallet(
            config=self.config,
            signals=self.signals,
            protowallet=protowallet,
            threading_parent=self.threading_manager,
            get_lang_code=self.language_chooser.get_current_lang_code,
        )

        qtprotowallet.signal_close_wallet.connect(
            lambda: self.close_tab(self.tab_wallets.indexOf(qtprotowallet.tab))
        )
        qtprotowallet.signal_create_wallet.connect(
            lambda: self.create_qtwallet_from_qtprotowallet(qtprotowallet)
        )

        # tutorial
        wallet_steps = WalletSteps(
            wallet_tabs=qtprotowallet.tabs,
            qtwalletbase=qtprotowallet,
        )
        if show_tutorial:
            protowallet.tutorial_index = 0
            wallet_steps.set_current_index(protowallet.tutorial_index)
            wallet_steps.set_visibilities()
        qtprotowallet.wallet_steps = wallet_steps

        tab_import_xpub = wallet_steps.tab_generators[TutorialStep.import_xpub]
        if not isinstance(tab_import_xpub, ImportXpubs):
            logger.error(f"{tab_import_xpub} is not of type ImportXpubs")
            return None

        def create_qtwallet_from_ui():
            if not isinstance(tab_import_xpub, ImportXpubs):
                logger.error(f"{tab_import_xpub} is not of type ImportXpubs")  # type: ignore[unreachable]
                return None

            if not tab_import_xpub.keystore_uis:
                Message("Cannot create wallet, because no keystores are available", type=MessageType.Error)
                return
            self.create_qtwallet_from_ui(
                wallet_tab=qtprotowallet.tab,
                protowallet=protowallet,
                keystore_uis=tab_import_xpub.keystore_uis,
            )

        wallet_steps.signal_create_wallet.connect(create_qtwallet_from_ui)

        # add to tabs
        self.tab_wallets.add_tab(
            tab=qtprotowallet.tab,
            icon=read_QIcon("file.png"),
            description=qtprotowallet.protowallet.id,
            focus=True,
            data=qtprotowallet,
        )

        return qtprotowallet

    def add_qt_wallet(
        self, wallet: Wallet, file_path: str | None = None, password: str | None = None
    ) -> QTWallet:
        def set_tab_widget_icon(tab: QWidget, icon: QIcon) -> None:
            idx = self.tab_wallets.indexOf(tab)
            if idx != -1:
                self.tab_wallets.setTabIcon(idx, icon)

        assert wallet.id not in self.qt_wallets, self.tr("A wallet with id {name} is already open.  ").format(
            name=wallet.id
        )

        with LoadingWalletTab(self.tab_wallets, wallet.id, focus=True):
            self.welcome_screen.remove_tab()
            qt_wallet = QTWallet(
                wallet,
                self.config,
                self.signals,
                self.mempool_data,
                self.fx,
                set_tab_widget_icon=set_tab_widget_icon,
                file_path=file_path,
                password=password,
                threading_parent=self.threading_manager,
                get_lang_code=self.language_chooser.get_current_lang_code,
            )

            # tutorial
            qt_wallet.wallet_steps = WalletSteps(
                wallet_tabs=qt_wallet.tabs,
                qtwalletbase=qt_wallet,
                qt_wallet=qt_wallet,
            )

        # add to tabs
        self.qt_wallets[wallet.id] = qt_wallet
        self.tab_wallets.add_tab(
            tab=qt_wallet.tab,
            icon=read_QIcon("status_waiting.svg"),
            description=qt_wallet.wallet.id,
            focus=True,
            data=qt_wallet,
        )

        # search_box = SearchWallets(
        #     lambda: list(self.qt_wallets.values()),
        #     signal_min=self.signals,
        #     parent=self.tab_wallets,
        # )
        # qt_wallet.tabs.set_top_right_widget(search_box)

        qt_wallet.wallet_steps.set_visibilities()
        self.language_chooser.add_signal_language_switch(
            self.signals.wallet_signals[qt_wallet.wallet.id].language_switch
        )
        self.signals.wallet_signals[qt_wallet.wallet.id].show_address.connect(self.show_address)
        self.signals.event_wallet_tab_added.emit()

        # this is a
        self.last_qtwallet = qt_wallet
        return qt_wallet

    def toggle_tutorial(self) -> None:
        qt_wallet = self.get_qt_wallet()
        if not qt_wallet:
            Message(self.tr("Please complete the wallet setup."))
            return

        if qt_wallet.wallet_steps:
            if qt_wallet.wallet.tutorial_index is None:
                qt_wallet.wallet.tutorial_index = qt_wallet.wallet_steps.step_bar.number_of_steps - 1
            else:
                qt_wallet.wallet.tutorial_index = None

            qt_wallet.wallet_steps.set_visibilities()

    def _get_qt_base_wallet(
        self,
        qt_base_wallets: Iterable[QtWalletBase],
        tab: QWidget | None = None,
        if_none_serve_last_active=False,
    ) -> Optional[QtWalletBase]:
        tab = self.tab_wallets.currentWidget() if tab is None else tab
        for qt_base_wallet in qt_base_wallets:
            if tab == qt_base_wallet.tab:
                return qt_base_wallet
        if if_none_serve_last_active:
            return self.last_qtwallet
        return None

    def get_qt_wallet(
        self, tab: QWidget | None = None, if_none_serve_last_active: bool = False
    ) -> Optional[QTWallet]:
        base_wallet = self._get_qt_base_wallet(
            self.qt_wallets.values(), tab=tab, if_none_serve_last_active=if_none_serve_last_active
        )
        if isinstance(base_wallet, QTWallet):
            return base_wallet
        return None

    def get_blockchain_of_any_wallet(self) -> Optional[bdk.Blockchain]:
        for qt_wallet in self.qt_wallets.values():
            if qt_wallet.wallet.blockchain:
                return qt_wallet.wallet.blockchain
        return None

    def show_address(self, addr: str, wallet_id: str, parent: QWidget | None = None) -> None:
        qt_wallet = self.qt_wallets.get(wallet_id)
        if not qt_wallet:
            return

        d = address_dialog.AddressDialog(
            self.fx,
            self.config,
            self.signals,
            qt_wallet.wallet,
            addr,
            self.mempool_data,
            parent=parent,
        )
        self.address_dialogs.append(d)
        d.show()

    def event_wallet_tab_closed(self) -> None:
        if not self.tab_wallets.count():
            self.welcome_screen.add_new_wallet_welcome_tab()

    def event_wallet_tab_added(self) -> None:
        pass

    def remove_qt_wallet(self, qt_wallet: Optional[QTWallet]) -> None:
        if not qt_wallet:
            return
        for i in range(self.tab_wallets.count()):
            if self.tab_wallets.widget(i) == qt_wallet.tab:
                self.tab_wallets.removeTab(i)

        self.add_recently_open_wallet(qt_wallet.file_path)

        qt_wallet.close()
        QTWallet.remove_lockfile(wallet_file_path=Path(qt_wallet.file_path))
        del self.qt_wallets[qt_wallet.wallet.id]
        self.event_wallet_tab_closed()

    def add_recently_open_wallet(self, file_path: str) -> None:
        self.config.add_recently_open_wallet(file_path)
        self.populate_recent_wallets_menu()

    def remove_all_qt_wallet(self) -> None:

        for qt_wallet in self.qt_wallets.copy().values():
            self.remove_qt_wallet(qt_wallet)

    def close_tab(self, index: int) -> None:
        self.tab_wallets.jump_to_last_active_tab()

        # qt_wallet
        qt_wallet = self.get_qt_wallet(tab=self.tab_wallets.widget(index))
        if qt_wallet:
            if not question_dialog(
                self.tr("Close wallet {id}?").format(id=qt_wallet.wallet.id), self.tr("Close wallet")
            ):
                return
            logger.info(self.tr("Closing wallet {id}").format(id=qt_wallet.wallet.id))
            self.save_qt_wallet(qt_wallet)
        else:
            logger.info(self.tr("Closing tab {name}").format(name=self.tab_wallets.tabText(index)))

        # get the tabdata before removing the tab
        tab_data = self.tab_wallets.tabData(index)
        self.tab_wallets.removeTab(index)
        if isinstance(tab_data, ThreadingManager):
            # this is necessary to ensure the closeevent
            # and with it the thread cleanup is called
            tab_data.end_threading_manager()

        if qt_wallet:
            self.remove_qt_wallet(qt_wallet)

        # other events
        self.event_wallet_tab_closed()

    def sync(self) -> None:
        qt_wallet = self.get_qt_wallet()
        if qt_wallet:
            qt_wallet.sync()

    def closeEvent(self, event: Optional[QCloseEvent]) -> None:
        self.config.last_wallet_files[str(self.config.network)] = [
            qt_wallet.file_path for qt_wallet in self.qt_wallets.values()
        ]

        self.write_current_open_txs_to_config()
        self.config.save()
        self.save_all_wallets()

        self.threading_manager.end_threading_manager()

        self.remove_all_qt_wallet()

        if self.new_startup_network:
            self.config.network = self.new_startup_network
            self.config.save()

        logger.info(f"Finished close handling of {self}")
        super().closeEvent(event)

    def restart(self, new_startup_network: bdk.Network | None = None) -> None:
        """
        Currently only works in Linux
        and then it seems that it freezes. So do not use

        Args:
            new_startup_network (bdk.Network | None, optional): _description_. Defaults to None.
        """
        args: List[str] = []  #  sys.argv[1:]
        self.new_startup_network = new_startup_network
        QCoreApplication.quit()

        status = QProcess.startDetached(sys.executable, ["-m", "bitcoin_safe"] + args)
        if not status:
            sys.exit(-1)

    def shutdown(self, new_startup_network: bdk.Network | None = None) -> None:
        self.new_startup_network = new_startup_network
        QCoreApplication.quit()

    def signal_handler(self, signum, frame) -> None:
        logger.info(f"Handling signal: {signum}")
        close_event = QCloseEvent()
        self.closeEvent(close_event)
        logger.info(f"Received signal {signum}, exiting.")
        QCoreApplication.quit()

    def setup_signal_handlers(self) -> None:
        for sig in [
            getattr(syssignal, attr)
            for attr in ["SIGTERM", "SIGINT", "SIGHUP", "SIGQUIT"]
            if hasattr(syssignal, attr)
        ]:
            syssignal.signal(sig, self.signal_handler)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from pathlib import Path
from typing import Optional

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont, QIcon, QPainter, QPixmap

from .util import create_button_box, read_QIcon

logger = logging.getLogger(__name__)
from PyQt6.QtGui import QAction, QFont, QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QDialog,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
)

from ...wallet import filename_clean


def question_dialog(
    text="", title="Question", buttons=QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Yes
) -> bool:
    msg_box = QMessageBox()
    msg_box.setWindowTitle(title)
    msg_box.setText(text)
    msg_box.setIcon(QMessageBox.Icon.Question)

    # Set the QDialogButtonBox as the standard button in the message box
    msg_box.setStandardButtons(buttons)

    # Execute the message box
    ret = msg_box.exec()

    # Check which button was clicked
    if ret == QMessageBox.StandardButton.Yes:
        return True
    elif ret == QMessageBox.StandardButton.No:
        return False
    elif ret == QMessageBox.StandardButton.Cancel:
        return False
    return False


class PasswordQuestion(QDialog):
    def __init__(self, parent=None, label_text=None) -> None:
        super(PasswordQuestion, self).__init__(parent)

        self.setWindowTitle(self.tr("Password Input"))
        self.setWindowIcon(read_QIcon("logo.svg"))

        self._layout = QVBoxLayout(self)

        label_text = label_text if label_text else self.tr("Please enter your password:")
        self.label = QLabel(label_text)
        self._layout.addWidget(self.label)

        self.password_input = QLineEdit(self)
        self.password_input.setEchoMode(QLineEdit.EchoMode.Password)
        self._layout.addWidget(self.password_input)

        # Create show/hide icons
        self.icon_show = create_icon_from_unicode("👁", size=18)
        self.icon_hide = create_icon_from_unicode("🙈", size=18)

        # Toggle password visibility action
        self.toggle_action = QAction(self.icon_show, self.tr("Show Password"), self)
        self.toggle_action.setFont(QFont("Arial", 12))  # Ensure Unicode support
        self.toggle_action.triggered.connect(self.toggle_password_visibility)
        self.password_input.addAction(self.toggle_action, QLineEdit.ActionPosition.TrailingPosition)

        self.submit_button = QPushButton(self.tr("Submit"), self)
        self.submit_button.clicked.connect(self.accept)
        self._layout.addWidget(self.submit_button)

    def toggle_password_visibility(self):
        if self.password_input.echoMode() == QLineEdit.EchoMode.Password:
            self.password_input.setEchoMode(QLineEdit.EchoMode.Normal)
            self.toggle_action.setIcon(self.icon_hide)
            self.toggle_action.setText(self.tr("Hide Password"))
        else:
            self.password_input.setEchoMode(QLineEdit.EchoMode.Password)
            self.toggle_action.setIcon(self.icon_show)
            self.toggle_action.setText(self.tr("Show Password"))

    def ask_for_password(self) -> Optional[str]:
        if self.exec() == QDialog.DialogCode.Accepted:
            return self.password_input.text()
        else:
            return None


def create_icon_from_unicode(unicode_char, font_name="Arial", size=18) -> QIcon:
    # Create a QPixmap object and set its size
    pixmap = QPixmap(32, 32)
    pixmap.fill(Qt.GlobalColor.transparent)

    # Create a QPainter object and draw text on the pixmap
    painter = QPainter(pixmap)
    painter.setFont(QFont(font_name, size))
    painter.drawText(0, 0, 32, 32, Qt.AlignmentFlag.AlignCenter, unicode_char)
    painter.end()

    # Create a QIcon object from the QPixmap
    return QIcon(pixmap)


class PasswordCreation(QDialog):
    def __init__(self, parent=None, window_title=None, label_text=None) -> None:
        super(PasswordCreation, self).__init__(parent)

        window_title = window_title if window_title else self.tr("Create Password")
        self.setWindowTitle(window_title)

        self._layout = QVBoxLayout(self)

        # First password input
        label_text = label_text if label_text else self.tr("Enter your password:")
        self.label1 = QLabel(label_text)
        self._layout.addWidget(self.label1)

        self.password_input1 = QLineEdit(self)
        self.password_input1.setEchoMode(QLineEdit.EchoMode.Password)
        self._layout.addWidget(self.password_input1)

        self.icon_show = create_icon_from_unicode("👁", size=18)
        self.icon_hide = create_icon_from_unicode("🙈", size=18)

        # Show password action for the first input
        self.show_password_action1 = QAction(create_icon_from_unicode("👁", size=18), self.tr("Show Password"))
        self.show_password_action1.setFont(
            QFont("Arial", 12)
        )  # Set the font to Arial to ensure Unicode support
        self.show_password_action1.triggered.connect(lambda: self.toggle_password_visibility())
        self.password_input1.addAction(self.show_password_action1, QLineEdit.ActionPosition.TrailingPosition)

        # Second password input
        self.label2 = QLabel(self.tr("Re-enter your password:"))
        self._layout.addWidget(self.label2)

        self.password_input2 = QLineEdit(self)
        self.password_input2.setEchoMode(QLineEdit.EchoMode.Password)
        self._layout.addWidget(self.password_input2)

        # Show password action for the second input
        # self.show_password_action2 = QAction(self.icon_show, "Show Password")
        # self.show_password_action2.triggered.connect(
        #     lambda: self.toggle_password_visibility(
        #         self.password_input2, self.show_password_action2
        #     )
        # )
        # self.password_input2.addAction(
        #     self.show_password_action2, QLineEdit.TrailingPosition
        # )

        # Submit button
        self.submit_button = QPushButton(self.tr("Submit"), self)
        self.submit_button.clicked.connect(self.verify_password)
        self._layout.addWidget(self.submit_button)

    def toggle_password_visibility(self) -> None:
        new_visibility = self.password_input1.echoMode() == QLineEdit.EchoMode.Password

        self._set_password_visibility(self.password_input1, self.show_password_action1, new_visibility)
        self._set_password_visibility(self.password_input2, self.show_password_action1, new_visibility)

    def _set_password_visibility(self, password_input, show_password_action, visibility) -> None:
        if visibility:
            password_input.setEchoMode(QLineEdit.EchoMode.Normal)
            show_password_action.setIcon(self.icon_hide)
            show_password_action.setToolTip(self.tr("Hide Password"))  # Set tooltip to "Hide Password"
        else:
            password_input.setEchoMode(QLineEdit.EchoMode.Password)
            show_password_action.setIcon(self.icon_show)
            show_password_action.setToolTip(self.tr("Show Password"))  # Set tooltip to "Show Password"

    def verify_password(self) -> None:
        # Check if passwords are identical
        password1 = self.password_input1.text()
        password2 = self.password_input2.text()

        if password1 == password2:
            self.accept()
        else:
            # Show a message box if passwords don't match
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Warning)
            msg_box.setText(self.tr("Passwords do not match!"))
            msg_box.setWindowTitle(self.tr("Error"))
            msg_box.setStandardButtons(QMessageBox.StandardButton.Ok)
            msg_box.exec()

    def get_password(self) -> Optional[str]:
        if self.exec() == QDialog.DialogCode.Accepted:
            return self.password_input1.text()
        else:
            return None


class WalletIdDialog(QDialog):
    def __init__(
        self, wallet_dir: Path, parent=None, window_title=None, label_text=None, prefilled=None
    ) -> None:
        super().__init__(parent)
        self.wallet_dir = wallet_dir
        window_title = window_title if window_title else self.tr("Choose wallet name")
        self.setWindowTitle(window_title)

        # Create layout
        layout = QVBoxLayout(self)

        # Add name label and input field
        label_text = label_text if label_text else self.tr("Wallet name:")
        self.name_label = QLabel(label_text)
        self.name_input = QLineEdit(prefilled if prefilled else "")
        layout.addWidget(self.name_label)
        layout.addWidget(self.name_input)

        # Add buttons
        self.buttonbox, self.buttons = create_button_box(self.check_wallet_existence, self.reject)
        layout.addWidget(self.buttonbox)

        # Set the layout
        self.name_input.setFocus()

    def check_wallet_existence(self) -> None:
        wallet_file = self.wallet_dir / self.filename
        if wallet_file.exists():
            QMessageBox.warning(
                self,
                self.tr("Error"),
                self.tr("The wallet {filename} exists already.").format(filename=wallet_file),
            )
        else:
            self.accept()  # Accept the dialog if wallet does not exist

    @property
    def wallet_id(self) -> str:
        return self.name_input.text()

    @property
    def filename(self) -> str:
        return filename_clean(self.wallet_id.lower())


if __name__ == "__main__":
    import sys

    from PyQt6.QtWidgets import QApplication

    app = QApplication(sys.argv)
    dialog = PasswordCreation()
    password = dialog.get_password()
    if password:
        print(f"Password created: {password}")
    sys.exit(app.exec())
    quit()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import enum
import logging
import math
import sys
from dataclasses import dataclass
from typing import Dict, Iterable, List, Tuple

from PyQt6.QtCore import QPointF, QRectF, pyqtSignal
from PyQt6.QtGui import (
    QColor,
    QLinearGradient,
    QMouseEvent,
    QPainter,
    QPainterPath,
    QPen,
)
from PyQt6.QtWidgets import QApplication, QTabWidget, QToolTip, QWidget

logger = logging.getLogger(__name__)


class FlowType(enum.Enum):
    InFlow = enum.auto()
    OutFlow = enum.auto()


@dataclass
class FlowIndex:
    flow_type: FlowType
    i: int

    def __hash__(self) -> int:
        return hash(tuple(self.__dict__.items()))


class SankeyWidget(QWidget):
    signal_on_label_click = pyqtSignal(FlowIndex)

    center_color = QColor("#7616ff")
    border_color = QColor("#7616ff")

    def __init__(self, show_tooltips=True, parent: QWidget | None = None) -> None:
        super().__init__(parent=parent)
        self.show_tooltips = show_tooltips

        self.colors: Dict[FlowIndex, QColor] = {}
        self.labels: Dict[FlowIndex, str] | None = None
        self.text_outline = True
        self.text_rects: List[Tuple[QRectF, FlowIndex]] = []

        self.node_width: float = 0
        self.x_offset: float = 0
        self.setMouseTracking(True)  # Enable mouse tracking

        # self.signal_on_label_click.connect(lambda flow_index: print(flow_index))

    @property
    def image_width(self) -> float:
        return self.width()

    @property
    def xscaling(self) -> float:
        return self.image_width / 1000

    @property
    def image_height(self) -> float:
        return min(self.height(), self.width() * 0.8)

    @property
    def yscaling(self) -> float:

        available_space_raw = max(self.sum_in_flows_raw / (1 - self.space_fraction), 1)

        return self.image_height / available_space_raw

    @property
    def flow_thickness(self) -> float:
        return self.sum_in_flows_raw * self.yscaling

    @property
    def in_flows(self) -> List[float]:
        return [width * self.yscaling for width in self.in_flows_raw]

    @property
    def out_flows(self) -> List[float]:
        return [width * self.yscaling for width in self.out_flows_raw]

    @property
    def end_in_y_positions(self) -> List[float]:
        return self.find_best_positions(
            self.in_flows, self.flow_thickness, offset=(self.image_height - self.flow_thickness) / 2
        )

    @property
    def end_out_y_positions(self) -> List[float]:
        return self.find_best_positions(
            self.out_flows, self.flow_thickness, offset=(self.image_height - self.flow_thickness) / 2
        )

    @property
    def in_flow_y_positions(self) -> List[float]:
        return self.find_best_positions(self.in_flows, self.image_height)

    @property
    def out_flow_y_positions(self) -> List[float]:
        return self.find_best_positions(self.out_flows, self.image_height)

    def set(
        self,
        in_flows: Iterable[float],
        out_flows: Iterable[float],
        colors: Dict[FlowIndex, QColor] | None = None,
        labels: Dict[FlowIndex, str] | None = None,
        tooltips: Dict[FlowIndex, str] | None = None,
        center_color=None,
        space_fraction=0.5,
        text_outline=True,
    ):
        self.center_color = center_color if center_color else self.center_color
        self.labels = labels if labels else {}
        self.tooltips = tooltips if tooltips else {}
        self.colors = colors if colors else {}
        self.text_outline = text_outline
        self.space_fraction = max(min(space_fraction, 0.95), 0)

        self.in_flows_raw = in_flows
        self.out_flows_raw = out_flows
        self.sum_in_flows_raw = sum(self.in_flows_raw)
        self.sum_out_flows_raw = sum(self.out_flows_raw)

        assert (
            self.sum_in_flows_raw == self.sum_out_flows_raw
        ), f"Inflows {self.sum_in_flows_raw} dont match outflows {self.sum_out_flows_raw}"

        self.node_width = 0
        self.x_offset = 0

    @staticmethod
    def _find_best_positions(thicknesses: List[float], available_space: float) -> List[float]:
        total_space = available_space - sum(thicknesses)
        space = total_space / max(len(thicknesses), 1)

        positions = []
        cursor = -space / 2
        for thickness in thicknesses:
            cursor += space + thickness / 2
            positions.append(cursor)
            # move the cursor further
            cursor += thickness / 2
        return positions

    @classmethod
    def find_best_positions(
        cls, thicknesses: List[float], available_space: float, offset: float = 0
    ) -> List[float]:
        positions = cls._find_best_positions(thicknesses, available_space=available_space)
        return [pos + offset for pos in positions]

    def _paint_one_side(
        self,
        painter: QPainter,
        flows: Iterable[float],
        y_start_positions: List[float],
        end_y_positions: List[float],
        flow_type: FlowType,
        reverse=False,
    ):
        image_left = (
            self.x_offset if not reverse else self.image_width + self.x_offset
        )  # Starting x position, adjusted if reversed
        image_right = self.image_width // 2  # End position of the bezier curve, depends on direction
        direction = (
            1 if not reverse else -1
        )  # Direction of bezier control points, reversed if flow is reversed

        # Draw flows
        for i, (start_y, end_y, width) in enumerate(zip(y_start_positions, end_y_positions, flows)):
            flow_index = FlowIndex(flow_type, i)
            start_x = image_left + (self.node_width + width // 2) * direction  # compensat for brush width
            end_x = image_right - width / 2 * direction  # compensat for brush width

            self.draw_path(
                painter,
                width,
                start_x,
                start_y,
                end_x,
                end_y,
                direction,
                self.colors.get(flow_index, self.border_color),
                self.center_color,
            )

            # Draw text at the start point
            if self.labels and flow_index in self.labels:
                self.draw_multiline_text(
                    painter,
                    self.labels[flow_index],
                    QPointF(image_left, start_y),
                    direction,
                    flow_index=FlowIndex(flow_type=flow_type, i=i),
                )

    def draw_path(
        self,
        painter: QPainter,
        width,
        start_x,
        start_y,
        end_x,
        end_y,
        direction: int,
        start_color: QColor,
        end_color: QColor,
    ):
        path = QPainterPath()
        path.moveTo(start_x, start_y)
        path.cubicTo(
            math.ceil(start_x + 100 * self.xscaling * direction),
            math.ceil(start_y),
            math.ceil(end_x - 100 * self.xscaling * direction),
            math.ceil(end_y),
            math.ceil(end_x) + direction,  # the +-1  is to close any gaps that might occur
            math.ceil(end_y),
        )

        gradient = QLinearGradient(QPointF(start_x, start_y), QPointF(end_x, end_y))
        gradient.setColorAt(0, start_color)
        gradient.setColorAt(1, end_color)
        pen = QPen()
        pen.setBrush(gradient)
        pen.setWidth(math.ceil(width))
        painter.setPen(pen)
        painter.drawPath(path)

    def draw_multiline_text(
        self, painter: QPainter, text: str, position: QPointF, direction: int, flow_index: FlowIndex
    ):
        painter.setPen(QColor("black"))
        font_metrics = painter.fontMetrics()
        lines = text.split("\n")  # Split the text into lines
        x, y = position.x(), position.y()

        for i, line in enumerate(lines):
            text_width = font_metrics.horizontalAdvance(line)
            # Adjust x position based on direction and width of the text
            text_x = x + (5 * direction) - (text_width if direction == -1 else 0)
            sub_position = QPointF(text_x, y + i * font_metrics.height())
            # save the full-text (not just the line)  at the sub_position in text_positions
            self.text_rects.append(
                (
                    QRectF(
                        sub_position.x(),
                        sub_position.y() - font_metrics.height(),
                        text_width,
                        font_metrics.height(),
                    ),
                    flow_index,
                )
            )
            # Draw text line by line
            if self.text_outline:
                self.draw_text_with_outline(painter, line, sub_position)
            else:
                painter.drawText(sub_position, line)

    def draw_text_with_outline(self, painter: QPainter, text: str, position: QPointF):
        # Configuration for the outline
        outline_offset = 1  # How far the outline is from the text
        outline_color = QColor("white")
        text_color = QColor("black")  # Color of the main text

        # Create a list of positions for the outline around the original position
        offsets = [
            QPointF(outline_offset, 0),
            QPointF(-outline_offset, 0),
            QPointF(0, outline_offset),
            QPointF(0, -outline_offset),
            QPointF(outline_offset, outline_offset),
            QPointF(-outline_offset, -outline_offset),
            QPointF(outline_offset, -outline_offset),
            QPointF(-outline_offset, outline_offset),
        ]

        # Draw the outline by offsetting the text slightly in various directions
        painter.setPen(outline_color)
        for offset in offsets:
            painter.drawText(position + offset, text)

        # Draw the main text on top
        painter.setPen(text_color)
        painter.drawText(position, text)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.text_rects.clear()

        self._paint_one_side(
            painter,
            self.in_flows,
            y_start_positions=self.in_flow_y_positions,
            end_y_positions=self.end_in_y_positions,
            flow_type=FlowType.InFlow,
        )
        self._paint_one_side(
            painter,
            self.out_flows,
            y_start_positions=self.out_flow_y_positions,
            end_y_positions=self.end_out_y_positions,
            reverse=True,
            flow_type=FlowType.OutFlow,
        )
        painter.end()

    def mouseMoveEvent(self, event: QMouseEvent | None) -> None:
        if not event:
            return
        if not self.show_tooltips:
            return
        for rect, flow_index in self.text_rects:
            if flow_index not in self.tooltips:
                continue
            if rect.contains(event.position()):
                # Convert widget-relative position to global position for the tooltip
                globalPos = self.mapToGlobal(event.position().toPoint())
                QToolTip.showText(globalPos, self.tooltips[flow_index], self)
                return  # Exit after showing one tooltip
        QToolTip.hideText()  # Hide tooltip if no text is hovered

    def mousePressEvent(self, event: QMouseEvent | None) -> None:
        if not event:
            return
        for rect, flow_index in self.text_rects:
            if rect.contains(event.position()):
                self.signal_on_label_click.emit(flow_index)
                break


if __name__ == "__main__":

    colors = {
        FlowIndex(FlowType.OutFlow, 1): QColor("#8af296"),
        FlowIndex(FlowType.OutFlow, 0): QColor("#f3f71b"),
        FlowIndex(FlowType.InFlow, 0): QColor("#8af296"),
    }
    # in_flows = [("apple", 50.0), ("banana", 30), ("orange", 20), ("lime", 10), ("blueberry", 40)]

    # out_flows = [
    #     ("fruit", 100.0),
    #     ("juice", 50.0),
    # ]

    in_flows = [
        70.0,
        30.0,
    ]

    out_flows = [
        65.0,
        30.0,
        5.0,
    ]
    labels = {
        FlowIndex(FlowType.InFlow, 0): "1\n1",
        FlowIndex(FlowType.InFlow, 1): "2",
        FlowIndex(FlowType.OutFlow, 0): "4",
        FlowIndex(FlowType.OutFlow, 1): "4",
        FlowIndex(FlowType.OutFlow, 2): "5",
    }

    app = QApplication(sys.argv)
    tabs = QTabWidget()

    sankey = SankeyWidget()
    sankey.set(in_flows=in_flows, out_flows=out_flows, colors=colors, text_outline=True, labels=labels)
    tabs.addTab(sankey, "sankey")
    tabs.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import List

from bitcoin_qr_tools.qr_widgets import QRCodeWidgetSVG
from PyQt6.QtCore import QMargins, Qt
from PyQt6.QtGui import QColor, QFont, QPalette, QResizeEvent, QWheelEvent
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QScrollArea,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.buttonedit import ButtonEdit
from bitcoin_safe.gui.qt.custom_edits import AnalyzerTextEdit


class TitledComponent(QWidget):
    def __init__(self, title, hex_color, parent=None) -> None:
        super().__init__(parent)

        self.title = QLabel(title, self)

        font = QFont()
        font.setBold(True)
        self.title.setFont(font)
        self.title.setAlignment(Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop)

        # Set the background color
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, QColor(hex_color))
        self.setPalette(palette)
        self.setAutoFillBackground(True)

        self._layout = QVBoxLayout(self)
        self._layout.setSpacing(3)

        self._layout.addWidget(self.title)


class ReceiveGroup(TitledComponent):
    def __init__(
        self, category: str, hex_color: str, address: str, qr_uri: str, width=170, parent=None
    ) -> None:
        super().__init__(title=category, hex_color=hex_color, parent=parent)
        self.setFixedWidth(width)

        # QR Code
        self.qr_code = QRCodeWidgetSVG(always_animate=True)
        self.qr_code.set_data_list([qr_uri])
        self._layout.addWidget(self.qr_code)

        input_field = AnalyzerTextEdit(address)
        self.text_edit = ButtonEdit(input_field=input_field)
        input_field.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.text_edit.setReadOnly(True)
        self.text_edit.add_copy_button()
        self.text_edit.input_field.setStyleSheet(
            f"""
            background-color: {hex_color};
            border: none;
            """
        )

        self.text_edit.setFixedHeight(60)
        self._layout.addWidget(self.text_edit)

    @property
    def address(self) -> str:
        return self.text_edit.input_field.text()

    @property
    def category(self) -> str:
        return self.title.text()


class NoVerticalScrollArea(QScrollArea):
    def __init__(self) -> None:
        super().__init__()
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        if scroll_bar := self.horizontalScrollBar():
            scroll_bar.valueChanged.connect(self.recenterVerticalScroll)

    def wheelEvent(self, event: QWheelEvent | None) -> None:
        # Override to do nothing, preventing vertical scrolling
        pass

    def recenterVerticalScroll(self) -> None:
        # Recenter the vertical scroll position when horizontal scrollbar state changes
        if self.widget() and (scroll_bar := self.verticalScrollBar()):
            max_scroll = scroll_bar.maximum()
            scroll_bar.setValue(max_scroll // 2)

    # Override resizeEvent to handle window resizing
    def resizeEvent(self, event: QResizeEvent | None) -> None:
        super().resizeEvent(event)
        self.recenterVerticalScroll()


class QuickReceive(QWidget):
    def __init__(self, title="Quick Receive", parent=None) -> None:
        super().__init__(parent=parent)

        self.setSizePolicy(
            QSizePolicy.Policy.Preferred,  # Horizontal size policy
            QSizePolicy.Policy.Fixed,  # Vertical size policy
        )

        # Horizontal Layout for Scroll Area content

        # Content Widget for the Scroll Area
        self.content_widget = QWidget()
        self.content_widget_layout = QHBoxLayout(self.content_widget)

        # Scroll Area
        self.scroll_area = NoVerticalScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.scroll_area.setWidget(self.content_widget)

        # Main Layout
        main_layout = QVBoxLayout(self)
        self.label_title = QLabel(title)
        font = QFont()
        font.setBold(True)
        self.label_title.setFont(font)
        main_layout.addWidget(self.label_title)
        main_layout.addWidget(self.scroll_area)

        # Group Box Management
        self.group_boxes: List[ReceiveGroup] = []

    def _qmargins_to_tuple(self, margins: QMargins) -> tuple[int, int, int, int]:
        return margins.left(), margins.top(), margins.right(), margins.bottom()

    def resizeEvent(self, event: QResizeEvent | None) -> None:
        for group_box in self.group_boxes:
            margins = self.content_widget_layout.getContentsMargins()
            scrollbar = self.scroll_area.horizontalScrollBar()
            group_box.setFixedHeight(
                self.height()
                - sum([m for m in margins if m])
                - sum(self._qmargins_to_tuple(self.scroll_area.contentsMargins()))
                - (scrollbar.height() if scrollbar else 0)
            )

    def add_box(self, receive_group: ReceiveGroup) -> None:
        self.group_boxes.append(receive_group)
        self.content_widget_layout.addWidget(receive_group)
        self.content_widget.adjustSize()

    def remove_box(self) -> None:
        if self.group_boxes:
            group_box = self.group_boxes.pop()
            group_box.setParent(None)  # type: ignore[call-overload]
            group_box.deleteLater()
            self.content_widget.adjustSize()

    def clear_boxes(self) -> None:
        while self.group_boxes:
            self.remove_box()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_safe.fx import FX
from bitcoin_safe.gui.qt.notification_bar import NotificationBar
from bitcoin_safe.gui.qt.util import icon_path
from bitcoin_safe.html import html_f, link
from bitcoin_safe.psbt_util import FeeInfo

from ...config import FEE_RATIO_HIGH_WARNING, NO_FEE_WARNING_BELOW, UserConfig

logger = logging.getLogger(__name__)

from typing import Optional

import bdkpython as bdk
from PyQt6.QtCore import QObject, Qt, pyqtSignal
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (
    QDoubleSpinBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from ...mempool import MempoolData, TxPrio
from ...signals import pyqtSignal
from ...util import Satoshis, format_fee_rate, unit_fee_str
from .block_buttons import (
    BaseBlock,
    ConfirmedBlock,
    MempoolButtons,
    MempoolProjectedBlock,
)


class FeeWarningBar(NotificationBar):
    def __init__(self) -> None:
        super().__init__(
            text="",
            optional_button_text="",
            has_close_button=False,
        )
        self.set_background_color("#FFDF00")
        self.set_icon(QIcon(icon_path("warning.png")))

        self.optionalButton.setVisible(False)

        self.setVisible(False)

    def setText(self, value: Optional[str]):
        self.textLabel.setText(value if value else "")


class FeeGroup(QObject):
    signal_set_fee_rate = pyqtSignal(float)

    def __init__(
        self,
        mempool_data: MempoolData,
        fx: FX,
        config: UserConfig,
        fee_info: FeeInfo | None = None,
        allow_edit=True,
        is_viewer=False,
        confirmation_time: bdk.BlockTime | None = None,
        url: str | None = None,
        fee_rate: float | None = None,
    ) -> None:
        super().__init__()

        self.fx = fx
        self.allow_edit = allow_edit
        self.config = config
        self.fee_info = fee_info

        fee_rate = fee_rate if fee_rate else (mempool_data.get_prio_fee_rates()[TxPrio.low])

        # add the groupBox_Fee
        self.groupBox_Fee = QGroupBox()
        self.groupBox_Fee_layout = QVBoxLayout(self.groupBox_Fee)
        self.groupBox_Fee.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding)
        self.groupBox_Fee.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.groupBox_Fee_layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self.groupBox_Fee_layout.setContentsMargins(0, 0, 0, 0)

        self._confirmed_block = None
        self._mempool_projected_block = None
        self._mempool_buttons = None

        self.high_fee_rate_warning_label = FeeWarningBar()
        self.high_fee_rate_warning_label.setHidden(True)
        self.groupBox_Fee_layout.addWidget(
            self.high_fee_rate_warning_label, alignment=Qt.AlignmentFlag.AlignHCenter
        )

        self.high_fee_warning_label = FeeWarningBar()
        self.high_fee_warning_label.setHidden(True)
        self.groupBox_Fee_layout.addWidget(
            self.high_fee_warning_label, alignment=Qt.AlignmentFlag.AlignHCenter
        )

        if confirmation_time:
            self._confirmed_block = ConfirmedBlock(
                mempool_data=mempool_data,
                url=url,
                confirmation_time=confirmation_time,
                fee_rate=fee_rate,
            )
        elif is_viewer:
            self._mempool_projected_block = MempoolProjectedBlock(
                mempool_data, config=self.config, fee_rate=fee_rate
            )
        else:
            self._mempool_buttons = MempoolButtons(mempool_data, max_button_count=3)

        if allow_edit:
            self.mempool().signal_click.connect(self.set_fee_rate)
        self.groupBox_Fee_layout.addWidget(
            self.mempool().button_group, alignment=Qt.AlignmentFlag.AlignHCenter
        )

        self.approximate_fee_label = QLabel()
        self.approximate_fee_label.setVisible(False)
        self.groupBox_Fee_layout.addWidget(
            self.approximate_fee_label, alignment=Qt.AlignmentFlag.AlignHCenter
        )

        self.rbf_fee_label = QLabel()
        self.rbf_fee_label.setWordWrap(True)
        self.rbf_fee_label.setHidden(True)
        self.groupBox_Fee_layout.addWidget(self.rbf_fee_label, alignment=Qt.AlignmentFlag.AlignHCenter)

        self.widget_around_spin_box = QWidget()
        self.widget_around_spin_box_layout = QHBoxLayout(self.widget_around_spin_box)
        self.widget_around_spin_box_layout.setContentsMargins(0, 0, 0, 0)  # Remove margins
        self.groupBox_Fee_layout.addWidget(
            self.widget_around_spin_box, alignment=Qt.AlignmentFlag.AlignHCenter
        )

        self.spin_fee_rate = QDoubleSpinBox()
        self.spin_fee_rate.setReadOnly(not allow_edit)
        self.spin_fee_rate.setSingleStep(1)  # Set the step size
        self.spin_fee_rate.setDecimals(1)  # Set the number of decimal places
        self.spin_fee_rate.setMaximumWidth(55)
        if fee_rate:
            self.spin_fee_rate.setValue(fee_rate)
        self.update_spin_fee_range()

        self.widget_around_spin_box_layout.addWidget(self.spin_fee_rate)

        self.spin_label = QLabel()
        self.spin_label.setText(unit_fee_str(self.config.network))
        self.widget_around_spin_box_layout.addWidget(self.spin_label)

        self.fee_amount_label = QLabel()
        self.fee_amount_label.setHidden(True)
        self.fee_amount_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.groupBox_Fee_layout.addWidget(self.fee_amount_label, alignment=Qt.AlignmentFlag.AlignHCenter)
        self.fiat_fee_label = QLabel()
        self.fiat_fee_label.setHidden(True)
        self.fiat_fee_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.groupBox_Fee_layout.addWidget(self.fiat_fee_label, alignment=Qt.AlignmentFlag.AlignHCenter)
        self.fx.signal_data_updated.connect(self.updateUi)

        self.label_block_number = QLabel()
        self.label_block_number.setHidden(True)
        self.groupBox_Fee_layout.addWidget(self.label_block_number, alignment=Qt.AlignmentFlag.AlignHCenter)

        if allow_edit:
            # self.spin_fee_rate.editingFinished.connect(lambda: self.set_fee_rate(self.spin_fee_rate.value()))
            self.spin_fee_rate.valueChanged.connect(lambda: self.set_fee_rate(self.spin_fee_rate.value()))
        self.mempool().mempool_data.signal_data_updated.connect(self.updateUi)
        self.mempool().refresh()
        self.updateUi()

    def mempool(self) -> BaseBlock:
        if self._confirmed_block:
            return self._confirmed_block
        if self._mempool_projected_block:
            return self._mempool_projected_block
        if self._mempool_buttons:
            return self._mempool_buttons
        raise Exception(f"{self.__class__.__name__} wasnt initialized correctly")

    def updateUi(self) -> None:
        self.groupBox_Fee.setTitle(self.tr("Fee"))
        self.rbf_fee_label.setText(
            html_f(self.tr("... is the minimum to replace the existing transactions."), bf=True)
        )
        self.approximate_fee_label.setText(html_f(self.tr("Approximate fee rate"), bf=True))

        # only in editor mode
        self.label_block_number.setHidden(self.spin_fee_rate.isReadOnly())
        if self.spin_fee_rate.value():
            self.label_block_number.setText(
                self.tr("in ~{n}. Block").format(
                    n=self.mempool().mempool_data.fee_rate_to_projected_block_index(
                        self.spin_fee_rate.value()
                    )
                    + 1
                )
            )

        self.approximate_fee_label.setVisible(self.fee_info.is_estimated if self.fee_info else False)
        if self.fee_info:
            self.approximate_fee_label.setToolTip(
                f'<html><body>The {"approximate " if   self.fee_info.is_estimated else "" }fee is {Satoshis( self.fee_info.fee_amount  , self.config.network).str_with_unit()}</body></html>'
            )

        self.set_fiat_fee_label()
        self.set_fee_amount_label()
        self.update_fee_rate_warning()
        self.mempool().refresh()

    def set_fiat_fee_label(self) -> None:
        self.fiat_fee_label.setHidden(self.fee_info is None)
        if self.fee_info is None:
            return

        fee = self.fee_info.vsize * self.spin_fee_rate.value()
        dollar_amount = self.fx.to_fiat("usd", int(fee))
        if dollar_amount is None:
            self.fiat_fee_label.setHidden(True)
            return

        dollar_text = self.fx.format_dollar(dollar_amount)
        if dollar_amount > 100:
            # make red when dollar amount high
            dollar_text = html_f(dollar_text, bf=True, color="red")
        self.fiat_fee_label.setText(dollar_text)

    def set_rbf_label(self, min_fee_rate: Optional[float]) -> None:
        self.rbf_fee_label.setVisible(bool(min_fee_rate))
        if min_fee_rate:
            self.rbf_fee_label.setText(
                (
                    self.tr("{rate} is the minimum for {rbf}").format(
                        rate=format_fee_rate(min_fee_rate, self.config.network),
                        rbf=link("https://github.com/bitcoin/bips/blob/master/bip-0125.mediawiki", "RBF"),
                    )
                )
            )
            self.rbf_fee_label.setTextFormat(Qt.TextFormat.RichText)
            self.rbf_fee_label.setOpenExternalLinks(True)  # Enable opening links

    def update_fee_rate_warning(self) -> None:
        fee_rate = self.spin_fee_rate.value()

        too_high = fee_rate > self.mempool().mempool_data.max_reasonable_fee_rate()
        if fee_rate <= NO_FEE_WARNING_BELOW:
            too_high = False

        self.high_fee_rate_warning_label.setVisible(too_high)
        if too_high:
            self.high_fee_rate_warning_label.setText(html_f(self.tr("High fee rate!"), bf=True))
            self.high_fee_rate_warning_label.setToolTip(
                self.tr("The high prio mempool fee rate is {rate}").format(
                    rate=format_fee_rate(
                        self.mempool().mempool_data.max_reasonable_fee_rate(), self.config.network
                    )
                )
            )

    def set_fee_to_send_ratio(
        self,
        fee_info: FeeInfo,
        total_non_change_output_amount: int,
        network: bdk.Network,
        force_show_fee_warning_on_0_amont=False,
    ) -> None:
        if total_non_change_output_amount <= 0:
            # the == 0 case is relevant
            self.high_fee_warning_label.setVisible(force_show_fee_warning_on_0_amont)
            self.high_fee_warning_label.setText(
                self.tr("{sent} is sent!").format(
                    sent=Satoshis(total_non_change_output_amount, network=network).str_with_unit()
                )
            )
            self.high_fee_warning_label.setToolTip(
                html_f(
                    self.tr("The transaction fee is:\n{fee}, and {sent} is sent!").format(
                        fee=Satoshis(fee_info.fee_amount, network).str_with_unit(),
                        sent=Satoshis(total_non_change_output_amount, network=network).str_with_unit(),
                    ),
                    add_html_and_body=True,
                )
            )
            return

        too_high = fee_info.fee_amount / total_non_change_output_amount > FEE_RATIO_HIGH_WARNING
        self.high_fee_warning_label.setVisible(too_high and not self.mempool().confirmation_time)
        if too_high:
            self.high_fee_warning_label.setText(
                html_f(
                    self.tr("High fee ratio: {ratio}%").format(
                        ratio=round(fee_info.fee_amount / total_non_change_output_amount * 100)
                    ),
                    bf=True,
                )
            )
            s = (
                self.tr(
                    "The estimated transaction fee is:\n{fee}, which is {percent}% of\nthe sending value {sent}"
                )
                if fee_info.is_estimated
                else self.tr(
                    "The transaction fee is:\n{fee}, which is {percent}% of\nthe sending value {sent}"
                )
            )
            self.high_fee_warning_label.setToolTip(
                html_f(
                    s.format(
                        fee=Satoshis(fee_info.fee_amount, network).str_with_unit(),
                        percent=round(fee_info.fee_amount / total_non_change_output_amount * 100),
                        sent=Satoshis(total_non_change_output_amount, self.config.network).str_with_unit(),
                    ),
                    add_html_and_body=True,
                )
            )

    def set_fee_rate(
        self,
        fee_rate: float,
        fee_info: FeeInfo | None = None,
        url: str | None = None,
        confirmation_time: bdk.BlockTime | None = None,
        chain_height: int | None = None,
    ) -> None:
        # this has to be done first, because it will trigger signals
        # that will also set self.fee_amount from the spin edit
        self._set_spin_fee_value(fee_rate)
        self.fee_info = fee_info

        self.spin_fee_rate.setHidden(fee_rate is None)
        self.spin_label.setHidden(fee_rate is None)

        self.mempool().refresh(
            fee_rate=fee_rate,
            confirmation_time=confirmation_time,
            chain_height=chain_height,
        )

        if url:
            self.mempool().set_url(url)

        self.updateUi()
        self.signal_set_fee_rate.emit(fee_rate)

    def _set_spin_fee_value(self, value: float) -> None:
        self.update_spin_fee_range(value)
        self.spin_fee_rate.setValue(value)

    def set_fee_amount_label(self):
        self.fee_amount_label.setHidden(self.fee_info is None)
        if self.fee_info is None:
            return

        fee = self.fee_info.fee_amount
        self.fee_amount_label.setText(Satoshis(int(fee), self.config.network).str_with_unit())

    def update_spin_fee_range(self, value: float = 0) -> None:
        fee_range = self.config.fee_ranges[self.config.network].copy()
        fee_range[1] = max(
            fee_range[1],
            value,
            self.spin_fee_rate.value(),
            max(self.mempool().mempool_data.fee_rates_min_max(0)),
        )
        self.spin_fee_rate.setRange(*fee_range)
